from __future__ import unicode_literals

from unittest import TestCase

from pandagg.node.query import GeoBoundingBox, GeoDistance, GeoPolygone, GeoShape


class GeoQueriesTestCase(TestCase):
    def test_geo_bounding_box_clause(self):
        body = {
            "pin.location": {
                "top_left": {"lat": 40.73, "lon": -74.1},
                "bottom_right": {"lat": 40.01, "lon": -71.12},
            }
        }
        expected = {"geo_bounding_box": body}

        q1 = GeoBoundingBox(
            field="pin.location",
            top_left={"lat": 40.73, "lon": -74.1},
            bottom_right={"lat": 40.01, "lon": -71.12},
        )
        q2 = GeoBoundingBox(
            pin__location={
                "top_left": {"lat": 40.73, "lon": -74.1},
                "bottom_right": {"lat": 40.01, "lon": -71.12},
            }
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                'geo_bounding_box, field=pin.location, bottom_right={"lat": 40.01, "lon": -71.12}, top_left={"lat": 40.73, "lon": -74.1}',
            )

    def test_geo_polygone_clause(self):
        body = {"person.location": {"points": [[-70, 40], [-80, 30], [-90, 20]]}}
        expected = {"geo_polygon": body}

        q1 = GeoPolygone(
            field="person.location", points=[[-70, 40], [-80, 30], [-90, 20]]
        )
        q2 = GeoPolygone(person__location={"points": [[-70, 40], [-80, 30], [-90, 20]]})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                "geo_polygon, field=person.location, points=[[-70, 40], [-80, 30], [-90, 20]]",
            )

    def test_geo_distance_clause(self):
        body = {"distance": "12km", "pin.location": "drm3btev3e86"}
        expected = {"geo_distance": body}

        q = GeoDistance(pin__location="drm3btev3e86", distance="12km")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(q.line_repr(depth=None), "geo_distance, field=pin.location")

    def test_geo_shape(self):
        body = {
            "location": {
                "shape": {
                    "type": "envelope",
                    "coordinates": [[13.0, 53.0], [14.0, 52.0]],
                },
                "relation": "within",
            }
        }
        expected = {"geo_shape": body}

        q1 = GeoShape(
            field="location",
            relation="within",
            shape={"type": "envelope", "coordinates": [[13.0, 53.0], [14.0, 52.0]]},
        )
        q2 = GeoShape(
            location={
                "shape": {
                    "type": "envelope",
                    "coordinates": [[13.0, 53.0], [14.0, 52.0]],
                },
                "relation": "within",
            }
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                'geo_shape, field=location, relation="within", shape={"coordinates": [[13.0, 53.0], [14.0, 52.0]], "type": "envelope"}',
            )



#!/usr/bin/env python
# coding: utf-8

"""Module to extract
"""

import logging
from os.path import join
import simplejson
import pandas as pd
import numpy as np

from examples.imdb.conf import DATA_DIR, OUTPUT_FILE_NAME

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class NpEncoder(simplejson.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)


if __name__ == "__main__":

    logging.info("LOADING TABLES")
    # tables imports
    reader_kwargs = {
        "encoding": "utf-8",
        "sep": ",",
        "quotechar": '"',
        "escapechar": "\\",
    }
    movies = pd.read_csv(join(DATA_DIR, "movies.csv"), index_col="id", **reader_kwargs)
    movies_genres = pd.read_csv(join(DATA_DIR, "movies_genres.csv"), **reader_kwargs)
    movies_directors = pd.read_csv(
        join(DATA_DIR, "movies_directors.csv"), **reader_kwargs
    )
    directors = pd.read_csv(
        join(DATA_DIR, "directors.csv"), index_col="id", **reader_kwargs
    )
    director_genres = pd.read_csv(
        join(DATA_DIR, "directors_genres.csv"), **reader_kwargs
    )
    roles = pd.read_csv(join(DATA_DIR, "roles.csv"), **reader_kwargs)
    actors = pd.read_csv(join(DATA_DIR, "actors.csv"), index_col="id", **reader_kwargs)

    # actors
    logging.info("SERIALIZE ACTORS")
    actor_roles = pd.merge(actors, roles, left_index=True, right_on="actor_id")
    actor_roles["serialized_roles"] = actor_roles.apply(
        lambda x: {
            "actor_id": x.actor_id,
            "first_name": x.first_name,
            "last_name": x.last_name,
            "full_name": "%s %s" % (x.first_name, x.last_name),
            "gender": x.gender,
            "role": x.role,
        },
        axis=1,
    )
    movie_serialized_actors = actor_roles.groupby("movie_id").serialized_roles.apply(
        list
    )

    # directors
    logging.info("SERIALIZE DIRECTORS")
    directors_grouped_genres = pd.DataFrame(
        director_genres.groupby("director_id").genre.apply(list)
    )
    movie_directors_extended = pd.merge(
        movies_directors, directors, left_on="director_id", right_index=True
    )
    movie_directors_extended = pd.merge(
        movie_directors_extended,
        directors_grouped_genres,
        how="left",
        left_on="director_id",
        right_index=True,
    )
    movie_directors_extended["serialized_directors"] = movie_directors_extended.apply(
        lambda x: {
            "director_id": x.director_id,
            "first_name": x.first_name,
            "last_name": x.last_name,
            "full_name": "%s %s" % (x.first_name, x.last_name),
            "genres": x.genre,
        },
        axis=1,
    )
    movie_serialized_directors = pd.DataFrame(
        movie_directors_extended.groupby("movie_id").serialized_directors.apply(list)
    )

    # movie genres
    logging.info("SERIALIZE MOVIE GENRES")
    movie_serialized_genres = movies_genres.groupby("movie_id").genre.apply(list)

    # merge
    logging.info("MERGE DATASETS")
    enriched_movies = pd.merge(
        movies, movie_serialized_actors, how="left", left_index=True, right_index=True
    )
    enriched_movies = pd.merge(
        enriched_movies,
        movie_serialized_directors,
        how="left",
        left_index=True,
        right_index=True,
    )
    enriched_movies = pd.merge(
        enriched_movies,
        movie_serialized_genres,
        how="left",
        left_index=True,
        right_index=True,
    )

    enriched_movies["nb_directors"] = enriched_movies.serialized_directors.apply(
        lambda x: len(x) if isinstance(x, list) else 0
    )
    enriched_movies["nb_roles"] = enriched_movies.serialized_roles.apply(
        lambda x: len(x) if isinstance(x, list) else 0
    )

    serialized = enriched_movies.apply(
        lambda x: {
            "movie_id": x.name,
            "name": x.loc["name"],
            "year": x.year,
            "genres": x.genre,
            "roles": x.serialized_roles,
            "nb_roles": x.nb_roles,
            "directors": x.serialized_directors,
            "nb_directors": x.nb_directors,
            "rank": x.loc["rank"],
        },
        axis=1,
    )

    # write
    logging.info("WRITE SERIALIZED DOCUMENTS")
    with open(join(DATA_DIR, OUTPUT_FILE_NAME), "w") as f:
        for s in serialized:
            f.write(simplejson.dumps(s, ignore_nan=True, cls=NpEncoder) + "\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy

from builtins import str as text

from future.utils import iterkeys, iteritems

from pandagg.interactive.response import IResponse
from pandagg.node.aggs.abstract import UniqueBucketAgg, MetricAgg, ShadowRoot
from pandagg.node.aggs.bucket import Nested
from pandagg.tree.response import AggsResponseTree


class Response:
    def __init__(self, data, search):
        self.data = data
        self.__search = search

        self.took = data["took"]
        self.timed_out = data["timed_out"]
        self._shards = data["_shards"]
        self.hits = Hits(data["hits"])
        self.aggregations = Aggregations(
            data.get("aggregations", {}),
            aggs=self.__search._aggs,
            index=self.__search._index,
            query=self.__search._query,
            client=self.__search._using,
        )
        self.profile = data.get("profile")

    def __iter__(self):
        return iter(self.hits)

    @property
    def success(self):
        return (
            self._shards["total"] == self._shards["successful"] and not self.timed_out
        )

    def __len__(self):
        return len(self.hits)

    def __repr__(self):
        return (
            "<Response> took %dms, success: %s, total result %s, contains %s hits"
            % (self.took, self.success, self.hits._total_repr(), len(self.hits))
        )


class Hits:
    def __init__(self, hits):
        self.data = hits
        self.total = hits["total"]
        self.hits = [Hit(hit) for hit in hits.get("hits", [])]
        self.max_score = hits["max_score"]

    def __len__(self):
        return len(self.hits)

    def __iter__(self):
        return iter(self.hits)

    def _total_repr(self):
        if not isinstance(self.total, dict):
            return str(self.total)
        if self.total.get("relation") == "eq":
            return str(self.total["value"])
        if self.total.get("relation") == "gte":
            return ">=%d" % self.total["value"]
        raise ValueError("Invalid total %s" % self.total)

    def __repr__(self):
        if not isinstance(self.total, dict):
            total_repr = text(self.total)
        elif self.total.get("relation") == "eq":
            total_repr = text(self.total["value"])
        elif self.total.get("relation") == "gte":
            total_repr = ">%d" % self.total["value"]
        else:
            raise ValueError("Invalid total %s" % self.total)
        return "<Hits> total: %s, contains %d hits" % (total_repr, len(self.hits))


class Hit:
    def __init__(self, data):
        self.data = data
        self._source = data.get("_source")
        self._score = data.get("_score")
        self._id = data.get("_id")
        self._type = data.get("_type")
        self._index = data.get("_index")

    def __repr__(self):
        return "<Hit %s> score=%.2f" % (self._id, self._score)


class Aggregations:
    def __init__(self, data, aggs, query, index, client):
        self.data = data
        self.__aggs = aggs
        self.__index = index
        self.__query = query
        self.__client = client

    def keys(self):
        return self.data.keys()

    def get(self, key):
        return self.data[key]

    def _parse_group_by(
        self,
        response,
        row=None,
        agg_name=None,
        until=None,
        ancestors=None,
        row_as_tuple=False,
        with_single_bucket_groups=False,
    ):
        """Recursive parsing of succession of unique child bucket aggregations.

        Yields each row for which last bucket aggregation generated buckets.
        """
        if ancestors is None:
            ancestors = self.__aggs.ancestors(until, id_only=True)
        if not row:
            row = [] if row_as_tuple else {}
        agg_name = self.__aggs.root if agg_name is None else agg_name
        if agg_name in response:
            agg_node = self.__aggs.get(agg_name)
            for key, raw_bucket in agg_node.extract_buckets(response[agg_name]):
                sub_row = copy.copy(row)
                if (
                    not isinstance(agg_node, UniqueBucketAgg)
                    or with_single_bucket_groups
                ):
                    if row_as_tuple:
                        sub_row.append(key)
                    else:
                        sub_row[agg_name] = key
                if agg_name == until:
                    # end real yield
                    if row_as_tuple:
                        yield tuple(sub_row), raw_bucket
                    else:
                        yield sub_row, raw_bucket
                elif agg_name in ancestors:
                    # yield children
                    for child in self.__aggs.children(agg_name, id_only=False):
                        for nrow, nraw_bucket in self._parse_group_by(
                            row=sub_row,
                            response=raw_bucket,
                            agg_name=child.name,
                            until=until,
                            row_as_tuple=row_as_tuple,
                            ancestors=ancestors,
                        ):
                            yield nrow, nraw_bucket

    def _normalize_buckets(self, agg_response, agg_name=None):
        """Recursive function to parse aggregation response as a normalized entities.
        Each response bucket is represented as a dict with keys (key, level, value, children)::

            {
                "level": "owner.id",
                "key": 35,
                "value": 235,
                "children": [
                ]
            }
        """
        agg_name = agg_name or self.__aggs.root
        agg_node = self.__aggs.get(agg_name)
        agg_children = self.__aggs.children(agg_node.name, id_only=False)
        for key, raw_bucket in agg_node.extract_buckets(agg_response[agg_name]):
            result = {
                "level": agg_name,
                "key": key,
                "value": agg_node.extract_bucket_value(raw_bucket),
            }
            normalized_children = [
                normalized_child
                for child in agg_children
                for normalized_child in self._normalize_buckets(
                    agg_name=child.name, agg_response=raw_bucket
                )
            ]
            if normalized_children:
                result["children"] = normalized_children
            yield result

    def _grouping_agg(self, name=None):
        """Return aggregation node that used as grouping node.
        Note: in case there is only a nested aggregation below that node, groupby nested clause.
        """
        # if provided
        if name is not None:
            if name not in self.__aggs:
                raise ValueError("Cannot group by <%s>, agg node does not exist" % name)
            if not self.__aggs._is_eligible_grouping_node(name):
                raise ValueError(
                    "Cannot group by <%s>, not a valid grouping aggregation" % name
                )
            # if parent of single nested clause and nested_autocorrect
            node = self.__aggs.get(name)
            if self.__aggs.nested_autocorrect:
                children = self.__aggs.children(node.identifier, id_only=False)
                if len(children) == 1 and isinstance(children[0], Nested):
                    return children[0]
            return node

        if isinstance(self.__aggs.get(self.__aggs.root), ShadowRoot):
            return None
        name = self.__aggs.deepest_linear_bucket_agg
        if name is None:
            return None
        return self.__aggs.get(name)

    def to_tabular(
        self,
        index_orient=True,
        grouped_by=None,
        expand_columns=True,
        expand_sep="|",
        normalize=True,
        with_single_bucket_groups=False,
    ):
        """Build tabular view of ES response grouping levels (rows) until 'grouped_by' aggregation node included is
        reached, and using children aggregations of grouping level as values for each of generated groups (columns).

        Suppose an aggregation of this shape (A & B bucket aggregations)::

            A──> B──> C1
                 ├──> C2
                 └──> C3

        With grouped_by='B', breakdown ElasticSearch response (tree structure), into a tabular structure of this shape::

                                  C1     C2    C3
            A           B
            wood        blue      10     4     0
                        red       7      5     2
            steel       blue      1      9     0
                        red       23     4     2

        :param index_orient: if True, level-key samples are returned as tuples, else in a dictionnary
        :param grouped_by: name of the aggregation node used as last grouping level
        :param normalize: if True, normalize columns buckets
        :return: index, index_names, values
        """
        grouping_agg = self._grouping_agg(grouped_by)
        if grouping_agg is None:
            index_values = [(tuple() if index_orient else dict(), self.data)]
            index_names = []
        else:
            index_names = [
                a.name
                for a in self.__aggs.ancestors(
                    grouping_agg.name, id_only=False, from_root=True
                )
                + [grouping_agg]
                if not isinstance(a, UniqueBucketAgg) or with_single_bucket_groups
            ]
            index_values = list(
                self._parse_group_by(
                    response=self.data,
                    row_as_tuple=index_orient,
                    until=grouping_agg.name,
                    with_single_bucket_groups=with_single_bucket_groups,
                )
            )
            if not index_values:
                return [], []

        if index_orient:
            rows = {
                row_index: self._serialize_columns(
                    row_values,
                    normalize=normalize,
                    total_agg=grouping_agg,
                    expand_columns=expand_columns,
                    expand_sep=expand_sep,
                )
                for row_index, row_values in index_values
            }
        else:
            rows = [
                dict(
                    row_index,
                    **self._serialize_columns(
                        row_values,
                        normalize=normalize,
                        total_agg=grouping_agg,
                        expand_columns=expand_columns,
                        expand_sep=expand_sep,
                    )
                )
                for row_index, row_values in index_values
            ]
        return index_names, rows

    def _serialize_columns(
        self, row_data, normalize, expand_columns, expand_sep, total_agg=None
    ):
        # extract value (usually 'doc_count') of grouping agg node
        result = {}
        if total_agg is not None and not isinstance(total_agg, ShadowRoot):
            result[total_agg.VALUE_ATTRS[0]] = total_agg.extract_bucket_value(row_data)
            grouping_agg_children = self.__aggs.children(
                total_agg.identifier, id_only=False
            )
        else:
            grouping_agg_children = self.__aggs.children(
                self.__aggs.root, id_only=False
            )

        # extract values of children, one columns per child
        for child in grouping_agg_children:
            if isinstance(child, (UniqueBucketAgg, MetricAgg)):
                result[child.name] = child.extract_bucket_value(row_data[child.name])
            elif expand_columns:
                for key, bucket in child.extract_buckets(row_data[child.name]):
                    result[
                        "%s%s%s" % (child.name, expand_sep, key)
                    ] = child.extract_bucket_value(bucket)
            elif normalize:
                result[child.name] = next(
                    self._normalize_buckets(row_data, child.name), None
                )
            else:
                result[child.name] = row_data[child.name]
        return result

    def to_dataframe(
        self, grouped_by=None, normalize_children=True, with_single_bucket_groups=False
    ):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                'Using dataframe output format requires to install pandas. Please install "pandas" or '
                "use another output format."
            )
        index_names, rows = self.to_tabular(
            index_orient=True,
            grouped_by=grouped_by,
            normalize=normalize_children,
            with_single_bucket_groups=with_single_bucket_groups,
        )
        index, values = zip(*iteritems(rows))
        if not index:
            return pd.DataFrame()
        if len(index[0]) == 0:
            index = (None,) * len(index)
        else:
            index = pd.MultiIndex.from_tuples(index, names=index_names)
        return pd.DataFrame(index=index, data=values)

    def to_normalized(self):
        children = []
        for k in sorted(iterkeys(self.data)):
            for child in self._normalize_buckets(self.data, k):
                children.append(child)
        return {"level": "root", "key": None, "value": None, "children": children}

    def to_tree(self):
        return AggsResponseTree(aggs=self.__aggs, index=self.__index).parse(self.data)

    def to_interactive_tree(self):
        return IResponse(
            tree=self.to_tree(),
            index_name=self.__index,
            query=self.__query,
            client=self.__client,
            depth=1,
        )

    def serialize(self, output="tabular", **kwargs):
        """
        :param output: output format, one of "raw", "tree", "interactive_tree", "normalized", "tabular", "dataframe"
        :param kwargs: tabular serialization kwargs
        :return:
        """
        if output == "raw":
            return self.data
        elif output == "tree":
            return self.to_tree()
        elif output == "interactive_tree":
            return self.to_interactive_tree()
        elif output == "normalized":
            return self.to_normalized()
        elif output == "tabular":
            return self.to_tabular(**kwargs)
        elif output == "dataframe":
            return self.to_dataframe(**kwargs)
        else:
            raise NotImplementedError("Unkown %s output format." % output)

    def __repr__(self):
        if not self.keys():
            return "<Aggregations> empty"
        return "<Aggregations> %s" % list(map(text, self.keys()))



from __future__ import unicode_literals


class InvalidAggregation(Exception):
    """Wrong aggregation definition"""


class MappingError(Exception):
    """Basic Mapping Error"""

    pass


class AbsentMappingFieldError(MappingError):
    """Field is not present in mapping."""

    pass


class InvalidOperationMappingFieldError(MappingError):
    """Invalid aggregation type on this mapping field."""

    pass


class VersionIncompatibilityError(Exception):
    """Pandagg is not compatible with this ElasticSearch version.
    """

    pass



# ensure loading of all classes in metaclass
from .compound import *
from .full_text import *
from .geo import *
from .joining import *
from .shape import *
from .span import *
from .specialized import *
from .specialized_compound import *
from .term_level import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase

from pandagg.query import Nested, Term


class JoiningQueriesTestCase(TestCase):
    def test_nested(self):
        n1 = Nested(
            query=Term(field="some_nested_path.id", value=2), path="some_nested_path"
        )
        n2 = Nested(query={"term": {"some_nested_path.id": 2}}, path="some_nested_path")

        for i, n in enumerate((n1, n2)):
            self.assertEqual(
                n.to_dict(),
                {
                    "nested": {
                        "path": "some_nested_path",
                        "query": {"term": {"some_nested_path.id": {"value": 2}}},
                    }
                },
            )
            self.assertEqual(
                n.__str__(),
                """<Query>
nested, path="some_nested_path"
└── query
    └── term, field=some_nested_path.id, value=2
""",
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import json

from lighttree import TreeBasedObj
from pandagg.tree.mapping import Mapping
from pandagg.interactive._field_agg_factory import field_classes_per_name


class IMapping(TreeBasedObj):
    """Interactive wrapper upon mapping tree, allowing field navigation and quick access to single clause aggregations
    computation.
    """

    _REPR_NAME = "Mapping"
    _NODE_PATH_ATTR = "name"

    def __init__(self, *args, **kwargs):
        self._client = kwargs.pop("client", None)
        self._index = kwargs.pop("index", None)
        root_path = kwargs.pop("root_path", None)
        depth = kwargs.pop("depth", 1)
        initial_tree = kwargs.pop("initial_tree", None)
        tree = Mapping(*args, **kwargs)
        super(IMapping, self).__init__(
            tree=tree, root_path=root_path, depth=depth, initial_tree=initial_tree
        )
        # if we reached a leave, add aggregation capabilities based on reached mapping type
        self._set_agg_property_if_required()

    def _clone(self, nid, root_path, depth):
        return IMapping(
            self._tree.subtree(nid),
            client=self._client,
            root_path=root_path,
            depth=depth,
            initial_tree=self._initial_tree,
            index=self._index,
        )

    def _set_agg_property_if_required(self):
        if self._client is not None and not self._tree.children(self._tree.root):
            field_node = self._tree.get(self._tree.root)
            if field_node.KEY in field_classes_per_name:
                self.a = field_classes_per_name[field_node.KEY](
                    mapping_tree=self._initial_tree,
                    client=self._client,
                    field=self._initial_tree.node_path(field_node.identifier),
                    index=self._index,
                )

    def __call__(self, *args, **kwargs):
        print(
            json.dumps(
                self._tree.to_dict(), indent=2, sort_keys=True, separators=(",", ": "),
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lighttree import TreeBasedObj
from pandagg.tree.query.abstract import Query


class IResponse(TreeBasedObj):

    """Interactive aggregation response.
    """

    _NODE_PATH_ATTR = "attr_name"
    _COERCE_ATTR = True

    def __init__(
        self,
        tree,
        client=None,
        index_name=None,
        root_path=None,
        depth=None,
        initial_tree=None,
        query=None,
    ):
        self._client = client
        self._index_name = index_name
        self._query = Query(query)
        super(IResponse, self).__init__(
            tree=tree, root_path=root_path, depth=depth, initial_tree=initial_tree
        )

    def _clone(self, nid, root_path, depth):
        return IResponse(
            client=self._client,
            index_name=self._index_name,
            tree=self._tree.subtree(nid),
            root_path=root_path,
            depth=depth,
            initial_tree=self._initial_tree,
            query=self._query,
        )

    def get_bucket_filter(self):
        """Build filters to select documents belonging to that bucket"""
        bucket_filter = self._initial_tree.get_bucket_filter(self._tree.root)
        return self._query.query(bucket_filter).to_dict()

    def list_documents(self, **body):
        """Return ES aggregation query to list documents belonging to given bucket.
        :return:
        """
        filter_query = self.get_bucket_filter()
        if filter_query:
            body["query"] = filter_query
        return self._client.search(index=self._index_name, body=body)



from unittest import TestCase

from pandagg.node.aggs.abstract import AggNode


class AbstractAggNodesTestCase(TestCase):
    def test_abstract_agg_node(self):
        class CustomAgg(AggNode):
            KEY = "custom_type"
            VALUE_ATTRS = ["bucket_value_path"]
            # would mean this agg can be applied only on boolean fields
            WHITELISTED_MAPPING_TYPES = ["boolean"]
            BLACKLISTED_MAPPING_TYPES = None

            # depends on ElasticSearch aggregation handling, since this is a fake Aggregation this get_filter method
            # doesn't really make sense, just wrote one so that all abstract methods are implemented
            def get_filter(self, key):
                return {"exists": {"field": key}}

            # example for unique bucket agg
            def extract_buckets(self, response_value):
                yield (None, response_value)

        node = CustomAgg(name="custom_agg_name", custom_body={"stuff": 2})
        self.assertEqual(node.to_dict(), {"custom_type": {"custom_body": {"stuff": 2}}})

        node = CustomAgg(
            name="custom_agg_name", custom_body={"stuff": 2}, meta="meta_stuff"
        )
        self.assertEqual(
            node.to_dict(),
            {"custom_type": {"custom_body": {"stuff": 2}}, "meta": "meta_stuff"},
        )

        self.assertEqual(
            node.__str__(),
            u'<CustomAgg, name=custom_agg_name, type=custom_type, body={"custom_body": {"stuff": 2}}>',
        )

        # suppose this aggregation type provide buckets in the following format
        hypothetic_es_response_bucket = {"bucket_value_path": 43}
        self.assertEqual(node.extract_bucket_value(hypothetic_es_response_bucket), 43)

        self.assertEqual(node.valid_on_field_type("string"), False)
        self.assertEqual(node.valid_on_field_type("boolean"), True)

    def test_single_bucket_node(self):
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from pandagg.tree.query.abstract import Compound


class Bool(Compound):
    KEY = "bool"


class Boosting(Compound):
    KEY = "boosting"


class ConstantScore(Compound):
    KEY = "constant_score"


class DisMax(Compound):
    KEY = "dis_max"


class FunctionScore(Compound):
    KEY = "function_score"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from future.utils import python_2_unicode_compatible
from builtins import str as text

from six import add_metaclass

from pandagg.node._node import Node
from pandagg.utils import DslMeta, get_dsl_class


@add_metaclass(DslMeta)
class UnnamedField:
    _type_name = "field"
    KEY = None

    def __init__(self, **body):
        self.body = body.copy()

    def to_named_field(self, name, _subfield=False):
        return Field(name, self.KEY, _subfield=_subfield, **self.body)

    get_dsl_class = classmethod(get_dsl_class)


class UnnamedComplexField(UnnamedField):
    KEY = None

    def __init__(self, **body):
        properties = body.pop("properties", None)
        if properties and not isinstance(properties, dict):
            raise ValueError("Invalid properties %s" % properties)
        self.properties = properties or {}
        super(UnnamedComplexField, self).__init__(**body)


class ShadowRoot(UnnamedComplexField):
    KEY = "_"


class UnnamedRegularField(UnnamedField):
    KEY = None

    def __init__(self, **body):
        fields = body.pop("fields", None)
        if fields and not isinstance(fields, dict):
            raise ValueError("Invalid fields %s" % fields)
        self.fields = fields
        super(UnnamedRegularField, self).__init__(**body)


@python_2_unicode_compatible
class Field(Node):
    def __init__(self, name, key, **body):
        # name will be used for dynamic attribute access in mapping interactive tree
        self.name = name
        self.KEY = key

        self._subfield = body.pop("_subfield", False)

        # rest of body
        self._body = body
        super(Field, self).__init__()

    @property
    def _identifier_prefix(self):
        return self.name

    @property
    def body(self):
        b = self._body.copy()
        if self.KEY in ("object", "_"):
            return b
        b["type"] = self.KEY
        return b

    @property
    def _display_pattern(self):
        if self.KEY == "object":
            return " {%s}"
        if self.KEY == "nested":
            return " [%s]"
        if self._subfield:
            return "~ %s"
        return "  %s"

    def line_repr(self, depth, **kwargs):
        if self.KEY == "_":
            return "_"
        max_size = 60
        pad = max(max_size - 4 * depth - len(self.name), 4)
        return (
            "%s%s%s"
            % (self.name, " " * pad, self._display_pattern)
            % self.KEY.capitalize()
        )

    def __str__(self):
        return "<Mapping Field %s> of type %s:\n%s" % (
            text(self.name),
            text(self.KEY),
            text(json.dumps(self.body, indent=4)),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from future.utils import iteritems
from lighttree.exceptions import NotFoundNodeError

from pandagg.node.mapping.abstract import (
    Field,
    UnnamedField,
    UnnamedRegularField,
    UnnamedComplexField,
)


from pandagg.exceptions import (
    AbsentMappingFieldError,
    InvalidOperationMappingFieldError,
)
from pandagg.tree._tree import Tree


class Mapping(Tree):

    node_class = Field
    KEY = None

    def __init__(self, *args, **kwargs):
        super(Mapping, self).__init__()
        if not kwargs and not args:
            return
        if len(args) > 1:
            raise ValueError(
                "Invalid mapping declaration. Got:\n*args: %s\n**kwargs: %s"
                % (args, kwargs)
            )

        if args:
            arg = args[0]
            if arg is None:
                # Mapping(None)
                return
            if isinstance(arg, Mapping):
                # Mapping(Mapping())
                self.insert(arg)
                return
            if isinstance(arg, dict):
                # Mapping({"properties": {}})
                kwargs = arg.copy()
            else:
                raise ValueError(
                    "Wrong declaration: args %s, kwargs %s" % (args, kwargs)
                )

        # Mapping(dynamic=False, properties={...}}
        properties = kwargs.pop("properties", None)
        dynamic = kwargs.pop("dynamic", None)

        # root
        root_node = Field("_", "_", dynamic=dynamic)
        self.insert_node(root_node)
        if properties:
            self._insert(root_node.identifier, properties, False)

    def _insert(self, pid, el, is_subfield):
        if not isinstance(el, dict):
            raise ValueError("Wrong declaration, got %s" % el)
        for name, field in iteritems(el):
            if isinstance(field, dict):
                field = field.copy()
                field = UnnamedField.get_dsl_class(field.pop("type", "object"))(**field)
            if not isinstance(field, UnnamedField):
                raise ValueError("Unsupported type %s" % type(field))
            node = field.to_named_field(name, _subfield=is_subfield)
            self.insert_node(node, parent_id=pid)
            if isinstance(field, UnnamedComplexField) and field.properties:
                self._insert(node.identifier, field.properties, False)
            if isinstance(field, UnnamedRegularField) and field.fields:
                if is_subfield:
                    raise ValueError(
                        "Cannot insert subfields into a subfield on field %s" % name
                    )
                self._insert(node.identifier, field.fields, True)

    def __nonzero__(self):
        return not self.is_empty()

    __bool__ = __nonzero__

    def to_dict(self, from_=None, depth=None):
        if self.root is None:
            return None
        from_ = self.root if from_ is None else from_
        node = self.get(from_)
        children_queries = {}
        if depth is None or depth > 0:
            if depth is not None:
                depth -= 1
            for child_node in self.children(node.identifier, id_only=False):
                children_queries[child_node.name] = self.to_dict(
                    from_=child_node.identifier, depth=depth
                )
        serialized_node = node.body
        if children_queries:
            if node.KEY in ("_", "object", "nested"):
                serialized_node["properties"] = children_queries
            else:
                serialized_node["fields"] = children_queries
        return serialized_node

    def resolve_path_to_id(self, path):
        if path in self._nodes_map:
            return path
        nid = self.root
        names = path.split(".")
        for name in names:
            matching_children = [
                c for c in self.children(nid, id_only=False) if c.name == name
            ]
            if len(matching_children) != 1:
                return path
            nid = matching_children[0].identifier
        return nid

    def get(self, key):
        return super(Mapping, self).get(self.resolve_path_to_id(key))

    def validate_agg_node(self, agg_node, exc=True):
        """Ensure if node has field or path that it exists in mapping, and that required aggregation type
        if allowed on this kind of field.
        :param agg_node: AggNode you want to validate on this mapping
        :param exc: boolean, if set to True raise exception if invalid
        :rtype: boolean
        """
        if hasattr(agg_node, "path"):
            if agg_node.path is None:
                # reverse nested
                return True
            return self.resolve_path_to_id(agg_node.path) in self

        if not hasattr(agg_node, "field"):
            return True

        # TODO take into account flattened data type
        field = self.resolve_path_to_id(agg_node.field)
        if field not in self:
            if not exc:
                return False
            raise AbsentMappingFieldError(
                u"Agg of type <%s> on non-existing field <%s>."
                % (agg_node.KEY, agg_node.field)
            )

        field_type = self.mapping_type_of_field(field)
        if not agg_node.valid_on_field_type(field_type):
            if not exc:
                return False
            raise InvalidOperationMappingFieldError(
                u"Agg of type <%s> not possible on field of type <%s>."
                % (agg_node.KEY, field_type)
            )
        return True

    def mapping_type_of_field(self, field_path):
        try:
            return self.get(field_path).KEY
        except NotFoundNodeError:
            raise AbsentMappingFieldError(
                u"<%s field is not present in mapping>" % field_path
            )

    def nested_at_field(self, field_path):
        nesteds = self.list_nesteds_at_field(field_path)
        if nesteds:
            return nesteds[0]
        return None

    def list_nesteds_at_field(self, field_path):
        path_nid = self.resolve_path_to_id(field_path)
        # from deepest to highest
        return [
            self.node_path(nid)
            for nid in self.ancestors(path_nid) + [path_nid]
            if self.get(nid).KEY == "nested"
        ]

    def node_path(self, nid):
        return ".".join(
            [
                self.get(id_).name
                for id_ in self.ancestors(nid, from_root=True) + [nid]
                if id_ != self.root
            ]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from future.utils import iteritems, python_2_unicode_compatible
from lighttree.interactive import Obj

from pandagg.interactive.mapping import IMapping
from pandagg.search import Search


def discover(using, index="*"):
    """
    :param using: Elasticsearch client
    :param index: Comma-separated list or wildcard expression of index names used to limit the request.
    """
    indices = Indices()
    for index_name, index_detail in iteritems(using.indices.get(index=index)):
        indices[index_name] = Index(
            client=using,
            name=index_name,
            mapping=index_detail["mappings"],
            settings=index_detail["settings"],
            aliases=index_detail["aliases"],
        )
    return indices


# until Proper Index class is written


@python_2_unicode_compatible
class Index(object):
    def __init__(self, name, settings, mapping, aliases, client=None):
        super(Index, self).__init__()
        self.client = client
        self.name = name
        self.settings = settings
        self._mapping = mapping
        self.mapping = IMapping(mapping, client=client, index=name)
        self.aliases = aliases

    def search(self, nested_autocorrect=True):
        return Search(
            using=self.client,
            mapping=self._mapping,
            index=self.name,
            nested_autocorrect=nested_autocorrect,
        )

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<Index '%s'>" % self.name


class Indices(Obj):
    _COERCE_ATTR = True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Inspired by https://python-guide-pt-br.readthedocs.io/fr/latest/writing/logging.html#logging-in-a-library
# Set default logging handler to avoid "No handler found" warnings.

import logging

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""AbstractParentAgg aggregations:
https://www.elastic.co/guide/en/elasticsearch/reference/2.3/search-aggregations-pipeline.html
"""

from .aggs import AbstractParentAgg


class AvgBucket(AbstractParentAgg):
    KEY = "avg_bucket"


class Derivative(AbstractParentAgg):
    KEY = "derivative"


class MaxBucket(AbstractParentAgg):
    KEY = "max_bucket"


class MinBucket(AbstractParentAgg):
    KEY = "min_bucket"


class SumBucket(AbstractParentAgg):
    KEY = "sum_bucket"


class StatsBucket(AbstractParentAgg):
    KEY = "stats_bucket"


class ExtendedStatsBucket(AbstractParentAgg):
    KEY = "extended_stats_bucket"


class PercentilesBucket(AbstractParentAgg):
    KEY = "percentiles_bucket"


class MovingAvg(AbstractParentAgg):
    KEY = "moving_avg"


class CumulativeSum(AbstractParentAgg):
    KEY = "cumulative_sum"


class BucketScript(AbstractParentAgg):
    KEY = "bucket_script"


class BucketSelector(AbstractParentAgg):
    KEY = "bucket_selector"


class BucketSort(AbstractParentAgg):
    KEY = "bucket_sort"


class SerialDiff(AbstractParentAgg):
    KEY = "serial_diff"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from unittest import TestCase
from pandagg.utils import equal_queries, equal_search


class UtilsTestCase(TestCase):
    def test_equal(self):
        q1 = {"bool": {"must": [{"term": {"field_A": 1}}, {"term": {"field_B": 2}}]}}
        q2 = {"bool": {"must": [{"term": {"field_B": 2}}, {"term": {"field_A": 1}}]}}
        non_equal_q = {
            "bool": {"must": [{"term": {"field_B": 2}}, {"term": {"field_A": 123}}]}
        }
        self.assertTrue(equal_queries(q1, q2))
        self.assertFalse(equal_queries(q1, non_equal_q))

        self.assertTrue(
            equal_search(
                {
                    "query": q1,
                    "sort": ["title", {"category": {"order": "desc"}}, "_score"],
                },
                {
                    "query": q2,
                    "sort": ["title", {"category": {"order": "desc"}}, "_score"],
                },
            )
        )
        self.assertFalse(
            equal_search(
                {
                    "query": q1,
                    "sort": ["title", {"category": {"order": "desc"}}, "_score"],
                },
                {
                    "query": q2,
                    "sort": ["title", "_score", {"category": {"order": "desc"}}],
                },
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals


from pandagg.node.query.abstract import ParentParameterClause


class Filter(ParentParameterClause):
    KEY = "filter"
    MULTIPLE = True


class Must(ParentParameterClause):
    KEY = "must"
    MULTIPLE = True


class MustNot(ParentParameterClause):
    KEY = "must_not"
    MULTIPLE = True


class Negative(ParentParameterClause):
    KEY = "negative"
    MULTIPLE = False


class Organic(ParentParameterClause):
    KEY = "organic"
    MULTIPLE = False


class Positive(ParentParameterClause):
    KEY = "positive"
    MULTIPLE = False


class Queries(ParentParameterClause):
    KEY = "queries"
    MULTIPLE = True


class QueryP(ParentParameterClause):
    # different name to avoid confusion with Query "tree" class
    KEY = "query"
    MULTIPLE = False


class Should(ParentParameterClause):
    KEY = "should"
    MULTIPLE = True



from __future__ import unicode_literals

from pandagg.node.query.term_level import Range
from pandagg.query import Bool, Term, Boosting
from tests import PandaggTestCase


class CompoundQueriesTestCase(PandaggTestCase):
    def test_bool(self):
        # test all possibles definitions
        b1 = Bool(
            filter=Term(some_field=2),
            should=[Range(other={"gte": 1}), Term(some=3)],
            boost=1.2,
        )
        b2 = Bool(
            filter=[Term(some_field=2)],
            should=[Range(other={"gte": 1}), Term(some=3)],
            boost=1.2,
        )
        b3 = Bool(
            filter={"term": {"some_field": 2}},
            should=[{"range": {"other": {"gte": 1}}}, {"term": {"some": 3}}],
            boost=1.2,
        )
        b4 = Bool(
            filter=[{"term": {"some_field": 2}}],
            should=[{"range": {"other": {"gte": 1}}}, {"term": {"some": 3}}],
            boost=1.2,
        )

        for b in (b1, b2, b3, b4):
            self.assertQueryEqual(
                b.to_dict(),
                {
                    "bool": {
                        "boost": 1.2,
                        "filter": [{"term": {"some_field": {"value": 2}}}],
                        "should": [
                            {"range": {"other": {"gte": 1}}},
                            {"term": {"some": {"value": 3}}},
                        ],
                    }
                },
            )

    def test_boosting(self):
        b1 = Boosting(
            positive=Term(field="text", value="apple"),
            negative=Term(field="text", value="pie tart fruit crumble tree"),
            negative_boost=0.5,
        )

        b2 = Boosting(
            positive={"term": {"text": "apple"}},
            negative={"term": {"text": "pie tart fruit crumble tree"}},
            negative_boost=0.5,
        )

        for b in (b1, b2):
            self.assertQueryEqual(
                b.to_dict(),
                {
                    "boosting": {
                        "negative": {
                            "term": {
                                "text": {"value": "pie tart fruit crumble " "tree"}
                            }
                        },
                        "negative_boost": 0.5,
                        "positive": {"term": {"text": {"value": "apple"}}},
                    }
                },
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import json

from future.utils import python_2_unicode_compatible, iteritems, string_types

from pandagg.tree._tree import Tree
from pandagg.tree.mapping import Mapping

from pandagg.node.aggs.abstract import (
    BucketAggNode,
    AggNode,
    ShadowRoot,
)
from pandagg.node.aggs.bucket import Nested, ReverseNested, Terms
from pandagg.node.aggs.pipeline import BucketSelector, BucketSort


@python_2_unicode_compatible
class Aggs(Tree):
    r"""
    Combination of aggregation clauses. This class provides handful methods to build an aggregation (see
    :func:`~pandagg.tree.aggs.Aggs.aggs` and :func:`~pandagg.tree.aggs.Aggs.groupby`), and is used as well
    to parse aggregations response in handy formats.

    Mapping declaration is optional, but doing so validates aggregation validity and automatically handles missing
    nested clauses.

    All following syntaxes are identical:

    From a dict:

    >>> Aggs({"per_user":{"terms":{"field":"user"}}})

    Using shortcut declaration: first argument is the aggregation type, other arguments are aggregation body parameters:

    >>> Aggs('terms', name='per_user', field='user')

    Using DSL class:

    >>> from pandagg.aggs import Terms
    >>> Aggs(Terms('per_user', field='user'))

    Dict and DSL class syntaxes allow to provide multiple clauses aggregations:

    >>> Aggs({"per_user":{"terms":{"field":"user"}, "aggs": {"avg_age": {"avg": {"field": "age"}}}}})

    With is similar to:

    >>> from pandagg.aggs import Terms, Avg
    >>> Aggs(Terms('per_user', field='user', aggs=Avg('avg_age', field='age')))

    :Keyword Arguments:
        * *mapping* (``dict`` or ``pandagg.tree.mapping.Mapping``) --
          Mapping of requested indice(s). Providing it will validate aggregations validity, and add required nested
          clauses if missing.

        * *nested_autocorrect* (``bool``) --
          In case of missing nested clauses in aggregation, if True, automatically add missing nested clauses, else
          raise error.

        * remaining kwargs:
          Used as body in aggregation
    """

    node_class = AggNode
    _crafted_root_name = "root"

    def __init__(self, *args, **kwargs):
        self.mapping = Mapping(kwargs.pop("mapping", None))
        self.nested_autocorrect = kwargs.pop("nested_autocorrect", False)
        super(Aggs, self).__init__()
        if args or kwargs:
            self._fill(*args, **kwargs)

    def __nonzero__(self):
        return bool(self.to_dict())

    __bool__ = __nonzero__

    def _fill(self, *args, **kwargs):
        if not args or len(args) > 2:
            raise ValueError()
        if len(args) == 2:
            # Aggs("my_term_agg", "terms", field="some_field")
            agg_name, agg_type = args
            sub_aggs = kwargs.pop("aggs", None) or kwargs.pop("aggregations", None)
            agg_node = self.get_node_dsl_class(agg_type)(agg_name, **kwargs)
            self.insert_node(agg_node)
            if sub_aggs:
                self.insert(Aggs(sub_aggs), parent_id=agg_node.identifier)
            return

        arg = args[0]

        if isinstance(arg, string_types):
            # Aggs("classification_type") -> {"classification_type": {"terms": {"field": "classification_type"}}}
            self.insert(Terms(arg, field=arg, **kwargs))
            return

        if kwargs:
            raise ValueError("Unexpected kwargs %s" % kwargs)

        if isinstance(arg, dict):
            # Aggs({"my_term_agg": {"terms": {"field": "some_field"}}, "my_filters_agg": {"filters": {...}}})
            pid = None
            if len(arg.keys()) > 1:
                shadow_root = ShadowRoot()
                self.insert_node(shadow_root)
                pid = shadow_root.identifier
            for agg_name, v in iteritems(arg):
                v = v.copy()
                sub_aggs = v.pop("aggs", None) or v.pop("aggregations", None)
                if len(v.keys()) != 1:
                    raise ValueError("Invalid aggregation: %s" % v)
                agg_type, agg_body = v.copy().popitem()
                agg_node = self.get_node_dsl_class(agg_type)(agg_name, **agg_body)
                self.insert_node(agg_node, parent_id=pid)
                if sub_aggs:
                    self.insert(Aggs(sub_aggs), parent_id=agg_node.identifier)
            return

        # Aggs([Terms("my_term_agg", field="some_field"), Filters(...)])
        # Aggs(Terms("my_term_agg", field="some_field"))
        if not isinstance(arg, (tuple, list)):
            arg = (arg,)
        pid = None
        if len(arg) > 1:
            shadow_root = ShadowRoot()
            self.insert_node(shadow_root)
            pid = shadow_root.identifier
        for ar in arg:
            if not isinstance(ar, (AggNode, Aggs)):
                raise ValueError("Invalid type %s, %s" % (type(ar), ar))
            self.insert(ar, parent_id=pid)

    def _clone_init(self, deep=False):
        return Aggs(
            mapping=self.mapping.clone(deep=deep),
            nested_autocorrect=self.nested_autocorrect,
        )

    def _is_eligible_grouping_node(self, nid):
        """Return whether node can be used as grouping node."""
        node = self.get(nid)
        if not isinstance(node, BucketAggNode):
            return False
        # special aggregations not returning anything
        if isinstance(node, (BucketSelector, BucketSort)):
            return False
        return True

    @property
    def deepest_linear_bucket_agg(self):
        """
        Return deepest bucket aggregation node (pandagg.nodes.abstract.BucketAggNode) of that aggregation that
        neither has siblings, nor has an ancestor with siblings.
        """
        if not self.root or not self._is_eligible_grouping_node(self.root):
            return None
        last_bucket_agg_name = self.root
        children = [
            c
            for c in self.children(last_bucket_agg_name, id_only=False)
            if self._is_eligible_grouping_node(c.identifier)
        ]
        while len(children) == 1:
            last_agg = children[0]
            if not self._is_eligible_grouping_node(last_agg.identifier):
                break
            last_bucket_agg_name = last_agg.name
            children = [
                c
                for c in self.children(last_bucket_agg_name, id_only=False)
                if self._is_eligible_grouping_node(c.identifier)
            ]
        return last_bucket_agg_name

    def _validate_aggs_parent_id(self, pid):
        """
        If pid is not None, ensure that pid belongs to tree, and that it refers to a bucket aggregation.

        Else, if not provided, return deepest bucket aggregation if there is no ambiguity (linear aggregations).
        KO: non-ambiguous::
            A──> B──> C1
                 └──> C2
            raise error

        OK: non-ambiguous (linear)::

            A──> B──> C1
            return C1
        """
        if pid is not None:
            if not self._is_eligible_grouping_node(pid):
                raise ValueError("Node id <%s> is not a bucket aggregation." % pid)
            return pid
        leaves = self.leaves(id_only=False)
        # root
        if len(leaves) == 0:
            return None

        if len(leaves) > 1 or not isinstance(leaves[0], BucketAggNode):
            raise ValueError(
                "Declaration is ambiguous, you must declare the node id under which these "
                "aggregations should be placed."
            )
        return leaves[0].identifier

    def groupby(self, *args, **kwargs):
        r"""
        Arrange passed aggregations in vertical/nested manner, above or below another agg clause.

        Given the initial aggregation::

            A──> B
            └──> C

        If `insert_below` = 'A'::

            A──> new──> B
                  └──> C

        If `insert_above` = 'B'::

            A──> new──> B
            └──> C

        `by` argument accepts single occurrence or sequence of following formats:

        * string (for terms agg concise declaration)
        * regular Elasticsearch dict syntax
        * AggNode instance (for instance Terms, Filters etc)

        If `insert_below` nor `insert_above` is provided by will be placed between the the deepest linear
        bucket aggregation if there is no ambiguity, and its children::

            A──> B      : OK generates     A──> B ─> C ─> by

            A──> B      : KO, ambiguous, must precise either A, B or C
            └──> C


        Accepted all Aggs.__init__ syntaxes

        >>> Aggs()\
        >>> .groupby('terms', name='per_user_id', field='user_id')
        {"terms_on_my_field":{"terms":{"field":"some_field"}}}

        Passing a dict:

        >>> Aggs().groupby({"terms_on_my_field":{"terms":{"field":"some_field"}}})
        {"terms_on_my_field":{"terms":{"field":"some_field"}}}


        Using DSL class:

        >>> from pandagg.aggs import Terms
        >>> Aggs().groupby(Terms('terms_on_my_field', field='some_field'))
        {"terms_on_my_field":{"terms":{"field":"some_field"}}}

        Shortcut syntax for terms aggregation: creates a terms aggregation, using field as aggregation name

        >>> Aggs().groupby('some_field')
        {"some_field":{"terms":{"field":"some_field"}}}

        Using a Aggs object:

        >>> Aggs().groupby(Aggs('per_user_id', 'terms', field='user_id'))
        {"terms_on_my_field":{"terms":{"field":"some_field"}}}

        Accepted declarations for multiple aggregations:


        :Keyword Arguments:
            * *insert_below* (``string``) --
              Parent aggregation name under which these aggregations should be placed
            * *insert_above* (``string``) --
              Aggregation name above which these aggregations should be placed
            * *at_root* (``string``) --
              Insert aggregations at root of aggregation query

            * remaining kwargs:
              Used as body in aggregation

        :rtype: pandagg.aggs.Aggs
        """
        insert_below = kwargs.pop("insert_below", None)
        insert_above = kwargs.pop("insert_above", None)
        at_root = kwargs.pop("at_root", None)
        if (
            sum(
                (
                    insert_below is not None,
                    insert_above is not None,
                    at_root is not None,
                )
            )
            > 1
        ):
            raise ValueError(
                'Must define at most one of "insert_above" or "insert_below" or "at_root".'
            )

        new_agg = self.clone(with_tree=True)
        if at_root is not None:
            if not new_agg.is_empty():
                existing_root = new_agg.get(new_agg.root)
                if isinstance(existing_root, ShadowRoot):
                    insert_below = existing_root.identifier
                else:
                    insert_above = existing_root.identifier

        # groupby({}, {}), but not groupby("some_field", "terms", field="yolo")
        if len(args) > 2 or (len(args) > 1 and not isinstance(args[0], string_types)):
            if kwargs:
                raise ValueError(
                    "Kwargs not allowed when passing multiple aggregations in args."
                )
            inserted_aggs = [Aggs(arg) for arg in args]
        # groupby([{}, {}])
        elif len(args) == 1 and isinstance(args[0], (list, tuple)):
            if kwargs:
                raise ValueError(
                    "Kwargs not allowed when passing multiple aggregations in args."
                )
            inserted_aggs = [Aggs(arg) for arg in args[0]]
        # groupby({})
        # groupby(Terms())
        # groupby('terms', name='per_tag', field='tag')
        else:
            inserted_aggs = [Aggs(*args, **kwargs)]

        if insert_above is not None:
            parent = new_agg.parent(insert_above, id_only=False)
            # None if insert_above was root
            insert_below = parent.identifier if parent is not None else None
            insert_above_subtree = new_agg.drop_subtree(insert_above)
            # if isinstance(by, (list, tuple)):
            for inserted_agg in inserted_aggs:
                new_agg.insert(inserted_agg, parent_id=insert_below)
                insert_below = inserted_agg.deepest_linear_bucket_agg
            new_agg.insert_tree(parent_id=insert_below, new_tree=insert_above_subtree)
            return new_agg

        insert_below = self._validate_aggs_parent_id(insert_below)

        # empty initial tree
        if insert_below is None:
            insert_below_subtrees = []
        else:
            insert_below_subtrees = [
                new_agg.drop_subtree(c.identifier)
                for c in new_agg.children(insert_below, id_only=False)
            ]
        for inserted_agg in inserted_aggs:
            new_agg.insert(inserted_agg, parent_id=insert_below)
            insert_below = inserted_agg.deepest_linear_bucket_agg
        for st in insert_below_subtrees:
            new_agg.insert_tree(parent_id=insert_below, new_tree=st)
        return new_agg

    def aggs(self, *args, **kwargs):
        r"""
        Arrange passed aggregations "horizontally".

        Given the initial aggregation::

            A──> B
            └──> C

        If passing multiple aggregations with `insert_below` = 'A'::

            A──> B
            └──> C
            └──> new1
            └──> new2

        Note: those will be placed under the `insert_below` aggregation clause id if provided, else under the deepest
        linear bucket aggregation if there is no ambiguity:

        OK::

            A──> B ─> C ─> new

        KO::

            A──> B
            └──> C

        `args` accepts single occurrence or sequence of following formats:

        * string (for terms agg concise declaration)
        * regular Elasticsearch dict syntax
        * AggNode instance (for instance Terms, Filters etc)


        :Keyword Arguments:
            * *insert_below* (``string``) --
              Parent aggregation name under which these aggregations should be placed
            * *at_root* (``string``) --
              Insert aggregations at root of aggregation query

            * remaining kwargs:
              Used as body in aggregation

        :rtype: pandagg.aggs.Aggs
        """
        insert_below = kwargs.pop("insert_below", None)
        at_root = kwargs.pop("at_root", None)
        new_agg = self.clone(with_tree=True)
        if at_root is None:
            insert_below = self._validate_aggs_parent_id(insert_below)
        else:
            if insert_below is not None:
                raise ValueError(
                    "Should provide at most one of 'at_root' and 'insert_below'."
                )
            if not new_agg.is_empty():
                existing_root = new_agg.get(new_agg.root)
                if isinstance(existing_root, ShadowRoot):
                    insert_below = existing_root.identifier
                else:
                    new_root = ShadowRoot()
                    new_agg._insert_node_above(new_root, existing_root.identifier)
                    insert_below = new_root.identifier

        deserialized = Aggs(*args, **kwargs)
        deserialized_root = deserialized.get(deserialized.root)
        if isinstance(deserialized_root, ShadowRoot):
            new_agg.merge(deserialized, nid=insert_below)
        else:
            new_agg.insert(deserialized, parent_id=insert_below)
        return new_agg

    def to_dict(self, from_=None, depth=None, with_name=True):
        if self.root is None:
            return {}
        from_ = self.root if from_ is None else from_
        node = self.get(from_)
        children_queries = {}
        if depth is None or depth > 0:
            if depth is not None:
                depth -= 1
            for child_node in self.children(node.name, id_only=False):
                children_queries[child_node.name] = self.to_dict(
                    from_=child_node.name, depth=depth, with_name=False
                )
        if isinstance(node, ShadowRoot):
            return children_queries
        else:
            node_query_dict = node.to_dict()
            if children_queries:
                node_query_dict["aggs"] = children_queries
        if with_name:
            return {node.name: node_query_dict}
        return node_query_dict

    def applied_nested_path_at_node(self, nid):
        # from current node to root
        for id_ in [nid] + self.ancestors(nid):
            node = self.get(id_)
            if isinstance(node, (Nested, ReverseNested)):
                return node.path
        return None

    def _insert_tree_below(self, new_tree, parent_id, deep):
        nroot = new_tree.get(new_tree.root)
        if isinstance(nroot, ShadowRoot) and not self.is_empty():
            return super(Aggs, self).merge(new_tree, nid=parent_id, deep=deep)
        return super(Aggs, self)._insert_tree_below(
            new_tree, parent_id=parent_id, deep=deep
        )

    def _insert_node_below(self, node, parent_id):
        """If mapping is provided, nested aggregations are automatically applied.
        """
        # if aggregation node is explicitely nested or reverse nested aggregation, do not override, but validate
        if (
            isinstance(node, Nested)
            or isinstance(node, ReverseNested)
            or not self.mapping
            or not hasattr(node, "field")
        ):
            return super(Aggs, self)._insert_node_below(node, parent_id)

        self.mapping.validate_agg_node(node)

        # from deepest to highest
        required_nested_level = self.mapping.nested_at_field(node.field)

        if self.is_empty():
            current_nested_level = None
        else:
            current_nested_level = self.applied_nested_path_at_node(parent_id)
        if current_nested_level == required_nested_level:
            return super(Aggs, self)._insert_node_below(node, parent_id)
        if not self.nested_autocorrect:
            raise ValueError(
                "Invalid %s agg on %s field. Invalid nested: expected %s, current %s."
                % (node.KEY, node.field, required_nested_level, current_nested_level)
            )
        if current_nested_level and (
            required_nested_level or "" in current_nested_level
        ):
            # requires reverse-nested
            # check if already exists in direct children, else create it
            child_reverse_nested = next(
                (
                    n
                    for n in self.children(parent_id, id_only=False)
                    if isinstance(n, ReverseNested) and n.path == required_nested_level
                ),
                None,
            )
            if child_reverse_nested:
                return super(Aggs, self)._insert_node_below(
                    node, child_reverse_nested.identifier
                )
            else:
                rv_node = ReverseNested(name="reverse_nested_below_%s" % parent_id)
                super(Aggs, self).insert_node(rv_node, parent_id)
                return super(Aggs, self)._insert_node_below(node, rv_node.identifier)

        # requires nested - apply all required nested fields
        for nested_lvl in reversed(self.mapping.list_nesteds_at_field(node.field)):
            if current_nested_level != nested_lvl:
                # check if already exists in direct children, else create it
                child_nested = next(
                    (
                        n
                        for n in (
                            self.children(parent_id, id_only=False)
                            if parent_id is not None
                            else []
                        )
                        if isinstance(n, Nested) and n.path == nested_lvl
                    ),
                    None,
                )
                if child_nested:
                    parent_id = child_nested.identifier
                    continue
                nested_node_name = (
                    "nested_below_root"
                    if parent_id is None
                    else "nested_below_%s" % parent_id
                )
                nested_node = Nested(name=nested_node_name, path=nested_lvl)
                super(Aggs, self)._insert_node_below(nested_node, parent_id)
                parent_id = nested_node.identifier
        super(Aggs, self)._insert_node_below(node, parent_id)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)


class AbstractLeafAgg(Aggs):
    KEY = None

    """Allow following syntax::

    >>> a = Avg("my_terms_agg", field="yolo")
    """

    def __init__(self, *args, **kwargs):
        mapping = kwargs.pop("mapping", None)
        nested_autocorrect = kwargs.pop("nested_autocorrect", None)
        super(AbstractLeafAgg, self).__init__(
            mapping=mapping, nested_autocorrect=nested_autocorrect
        )
        node = self.get_node_dsl_class(self.KEY)(*args, **kwargs)
        self.insert_node(node)


class AbstractParentAgg(Aggs):
    KEY = None

    """Allow following syntax::

    >>> a = Terms("my_terms_agg", field="yolo", aggs={...})
    """

    def __init__(self, *args, **kwargs):
        mapping = kwargs.pop("mapping", None)
        nested_autocorrect = kwargs.pop("nested_autocorrect", None)
        aggs = kwargs.pop("aggs", None)
        super(AbstractParentAgg, self).__init__(
            mapping=mapping, nested_autocorrect=nested_autocorrect
        )
        node = self.get_node_dsl_class(self.KEY)(*args, **kwargs)
        self.insert_node(node)

        if aggs:
            if isinstance(aggs, dict):
                self.insert(Aggs(aggs), parent_id=node.identifier)
                return
            if not isinstance(aggs, (list, tuple)):
                aggs = (aggs,)
            for agg in aggs:
                if not isinstance(agg, (AggNode, Aggs)):
                    raise ValueError("Unsupported type %s" % type(agg))
                self.insert(agg, node.identifier)



from .abstract import Leaf


class Exists(Leaf):
    KEY = "exists"


class Fuzzy(Leaf):
    KEY = "fuzzy"


class Ids(Leaf):
    KEY = "ids"


class Prefix(Leaf):
    KEY = "prefix"


class Range(Leaf):
    KEY = "range"


class Regexp(Leaf):
    KEY = "regexp"


class Term(Leaf):
    KEY = "term"


class Terms(Leaf):
    KEY = "terms"


class TermsSet(Leaf):
    KEY = "terms_set"


class Type(Leaf):
    KEY = "type"


class Wildcard(Leaf):
    KEY = "wildcard"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from mock import patch

from pandagg.node.query._parameter_clause import Must
from pandagg.node.query.joining import Nested
from pandagg.query import Query, Range, Prefix, Ids, Term, Terms, Nested
from pandagg.node.query.term_level import Term as TermNode, Exists as ExistsNode
from pandagg.node.query.joining import Nested as NestedNode
from pandagg.tree.query.compound import Bool
from pandagg.node.query.compound import Bool as BoolNode

from pandagg.utils import equal_queries, ordered
from tests import PandaggTestCase


class QueryTestCase(PandaggTestCase):
    def setUp(self):
        patcher = patch("uuid.uuid4", side_effect=range(1000))
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_term_query(self):
        q = Query(Term(field="some_field", value=2))
        self.assertEqual(q.to_dict(), {"term": {"some_field": {"value": 2}}})

    def test_query_compound(self):
        q = Query(Bool(filter=[Term(field="yolo", value=2)], boost=2))
        self.assertEqual(
            q.to_dict(),
            {"bool": {"boost": 2, "filter": [{"term": {"yolo": {"value": 2}}}]}},
        )
        self.assertEqual(
            q.__str__(),
            """<Query>
bool, boost=2
└── filter
    └── term, field=yolo, value=2
""",
        )

    def test_new_bool_must_above_node(self):
        initial_q = Query()
        result_q = initial_q.bool(_name="root_bool")
        self.assertEqual(result_q.to_dict(), None)

        # above element (in filter? in must? in should?)
        initial_q = Query(Term(field="some_field", value=2, _name="term_q"))
        result_q = initial_q.bool(_name="root_bool", child="term_q", child_param="must")
        self.assertEqual(
            result_q.to_dict(),
            {
                "bool": {
                    "_name": "root_bool",
                    "must": [{"term": {"_name": "term_q", "some_field": {"value": 2}}}],
                }
            },
        )

        # above element (without declaring above: default behavior is wrapping at root)
        initial_q = Query(Term(field="some_field", value=2, _name="term_q"))
        result_q = initial_q.bool(_name="root_bool", child_param="must")
        self.assertEqual(
            result_q.to_dict(),
            {
                "bool": {
                    "_name": "root_bool",
                    "must": [{"term": {"_name": "term_q", "some_field": {"value": 2}}}],
                }
            },
        )

        # above bool element
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )
        result_q = initial_q.bool(
            child="init_bool", _name="top_bool", filter=Range(field="price", gte=12)
        )
        self.assertEqual(
            ordered(result_q.to_dict()),
            ordered(
                {
                    "bool": {
                        "_name": "top_bool",
                        "filter": [{"range": {"price": {"gte": 12}}}],
                        "must": [
                            {
                                "bool": {
                                    "_name": "init_bool",
                                    "should": [
                                        {
                                            "term": {
                                                "_name": "prod_term",
                                                "some_field": {"value": "prod"},
                                            }
                                        },
                                        {
                                            "term": {
                                                "_name": "pizza_term",
                                                "other_field": {"value": "pizza"},
                                            }
                                        },
                                    ],
                                }
                            }
                        ],
                    }
                }
            ),
        )

        # above element in existing bool query
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )
        result_q = initial_q.bool(
            child="init_bool",
            _name="top_bool",
            child_param="must",
            filter=Range(field="price", gte=12),
        )
        self.assertTrue(
            ordered(result_q.to_dict()),
            ordered(
                {
                    "bool": {
                        "_name": "top_bool",
                        "filter": [{"range": {"price": {"gte": 12}}}],
                        "must": [
                            {
                                "bool": {
                                    "_name": "init_bool",
                                    "should": [
                                        {"term": {"some_field": {"value": "prod"}}},
                                        {"term": {"other_field": {"value": "pizza"}}},
                                    ],
                                }
                            }
                        ],
                    }
                }
            ),
        )

    def test_new_bool_below_node(self):
        # below single child parameter
        initial_q = Query(Nested(path="some_nested", _name="nested_id"))
        result_q = initial_q.bool(
            parent="nested_id", _name="bool_id", filter={"term": {"some_nested.id": 2}}
        )
        self.assertEqual(
            result_q.to_dict(),
            {
                "nested": {
                    "_name": "nested_id",
                    "path": "some_nested",
                    "query": {
                        "bool": {
                            "_name": "bool_id",
                            "filter": [{"term": {"some_nested.id": {"value": 2}}}],
                        }
                    },
                }
            },
        )

        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )
        result_q = initial_q.bool(
            parent="init_bool",
            parent_param="filter",
            _name="below_bool",
            must={"term": {"new_field": 2}},
        )
        self.assertTrue(
            equal_queries(
                result_q.to_dict(),
                {
                    "bool": {
                        "_name": "init_bool",
                        "filter": [
                            {
                                "bool": {
                                    "_name": "below_bool",
                                    "must": [{"term": {"new_field": {"value": 2}}}],
                                }
                            }
                        ],
                        "should": [
                            {
                                "term": {
                                    "_name": "prod_term",
                                    "some_field": {"value": "prod"},
                                }
                            },
                            {
                                "term": {
                                    "_name": "pizza_term",
                                    "other_field": {"value": "pizza"},
                                }
                            },
                        ],
                    }
                },
            )
        )

    def test_not_possible_parent_child(self):
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )

        # UNDER non-existing
        with self.assertRaises(ValueError) as e:
            initial_q.must(
                {"term": {"new_field": 2}},
                parent="not_existing_node",
                _name="somewhere",
            )
        self.assertEqual(
            e.exception.args,
            ("Parent <not_existing_node> does not exist in current query.",),
        )

        with self.assertRaises(ValueError) as e:
            initial_q.must(
                {"term": {"new_field": 2}},
                parent="not_existing_node",
                _name="somewhere",
            )
        self.assertEqual(
            e.exception.args,
            ("Parent <not_existing_node> does not exist in current query.",),
        )

        # ABOVE non-existing
        with self.assertRaises(ValueError) as e:
            initial_q.must(
                {"term": {"new_field": 2}},
                child="not_existing_node",
                _name="somewhere",
            )
        self.assertEqual(
            e.exception.args,
            ("Child <not_existing_node> does not exist in current query.",),
        )

        with self.assertRaises(ValueError) as e:
            initial_q.must(
                {"term": {"new_field": 2}},
                child="not_existing_node",
                _name="somewhere",
            )
        self.assertEqual(
            e.exception.args,
            ("Child <not_existing_node> does not exist in current query.",),
        )

        # UNDER leaf
        with self.assertRaises(Exception) as e:
            initial_q.must(
                {"term": {"new_field": 2}}, parent="pizza_term", _name="somewhere"
            )
        self.assertEqual(
            e.exception.args,
            (
                "Cannot place clause under non-compound clause <pizza_term> of type <term>.",
            ),
        )

        with self.assertRaises(Exception) as e:
            initial_q.must(
                {"term": {"new_field": 2}}, parent="pizza_term", _name="somewhere"
            )
        self.assertEqual(
            e.exception.args,
            (
                "Cannot place clause under non-compound clause <pizza_term> of type <term>.",
            ),
        )

    def test_replace_all_existing_bool(self):
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )
        result_q = initial_q.bool(
            _name="init_bool",
            filter=Range(field="price", gte=12, _name="price_range"),
            mode="replace_all",
        )
        self.assertNotIn("prod_term", result_q)
        self.assertNotIn("pizza_term", result_q)
        self.assertIn("price_range", result_q)

    def test_add_existing_bool(self):
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
            )
        )
        result_q = initial_q.bool(
            _name="init_bool",
            filter=Range(field="price", gte=12, _name="price_range"),
            should=Term(field="other_field", value="new_pizza", _name="new_pizza_term"),
            mode="add",
        )
        self.assertIn("prod_term", result_q)
        self.assertIn("pizza_term", result_q)
        self.assertIn("new_pizza_term", result_q)
        self.assertIn("price_range", result_q)
        self.assertQueryEqual(
            result_q.to_dict(),
            {
                "bool": {
                    "_name": "init_bool",
                    "filter": [
                        {"range": {"_name": "price_range", "price": {"gte": 12}}}
                    ],
                    "should": [
                        {
                            "term": {
                                "_name": "prod_term",
                                "some_field": {"value": "prod"},
                            }
                        },
                        {
                            "term": {
                                "_name": "pizza_term",
                                "other_field": {"value": "pizza"},
                            }
                        },
                        {
                            "term": {
                                "_name": "new_pizza_term",
                                "other_field": {"value": "new_pizza"},
                            }
                        },
                    ],
                }
            },
        )

    def test_replace_existing_bool(self):
        # replace != replace_all
        initial_q = Query(
            Bool(
                _name="init_bool",
                should=[
                    Term(field="some_field", value="prod", _name="prod_term"),
                    Term(field="other_field", value="pizza", _name="pizza_term"),
                ],
                filter=Range(field="price", gte=50, _name="very_high_price"),
                minimum_should_match=2,
            )
        )
        result_q = initial_q.bool(
            _name="init_bool",
            filter=Range(field="price", gte=20, _name="high_price"),
            mode="replace",
        )
        self.assertIn("prod_term", result_q)
        self.assertIn("pizza_term", result_q)
        self.assertIn("high_price", result_q)
        self.assertNotIn("very_high_price", result_q)
        self.assertTrue(
            equal_queries(
                result_q.to_dict(),
                {
                    "bool": {
                        "_name": "init_bool",
                        "filter": [
                            {"range": {"_name": "high_price", "price": {"gte": 20}}}
                        ],
                        "should": [
                            {
                                "term": {
                                    "_name": "prod_term",
                                    "some_field": {"value": "prod"},
                                }
                            },
                            {
                                "term": {
                                    "_name": "pizza_term",
                                    "other_field": {"value": "pizza"},
                                }
                            },
                        ],
                        "minimum_should_match": 2,
                    }
                },
            )
        )

    def test_must_at_root(self):
        q_i1 = Query()
        q1 = q_i1.must(
            Term(field="some_field", value=2, _name="term_nid"), _name="bool_nid",
        )
        self.assertEqual(len(q_i1.list()), 0)
        bool_ = next((c for c in q1.list() if isinstance(c, BoolNode)))
        self.assertEqual(bool_.name, "bool_nid")
        next((c for c in q1.list() if isinstance(c, Must)))
        term = next((c for c in q1.list() if isinstance(c, TermNode)))
        self.assertEqual(term.name, "term_nid")

        self.assertEqual(
            q1.__str__(),
            """<Query>
bool
└── must
    └── term, field=some_field, value=2
""",
        )

        q_i2 = Query()
        q2 = q_i2.must({"term": {"some_field": {"value": 2}}})
        self.assertEqual(len(q_i2.list()), 0)
        next((c for c in q2.list() if isinstance(c, BoolNode)))
        next((c for c in q2.list() if isinstance(c, Must)))
        next((c for c in q2.list() if isinstance(c, TermNode)))
        self.assertEqual(
            q2.__str__(),
            """<Query>
bool
└── must
    └── term, field=some_field, value=2
""",
        )

        # with multiple conditions (with different declarations)
        q_i3 = Query()
        q3 = q_i3.must(
            [
                {"exists": {"field": "some_field"}},
                {"term": {"other_field": {"value": 5}}},
            ]
        )
        next((c for c in q3.list() if isinstance(c, BoolNode)))
        next((c for c in q3.list() if isinstance(c, Must)))
        next((c for c in q3.list() if isinstance(c, TermNode)))
        next((c for c in q3.list() if isinstance(c, ExistsNode)))
        self.assertEqual(
            q3.__str__(),
            """<Query>
bool
└── must
    ├── exists, field=some_field
    └── term, field=other_field, value=5
""",
        )

    def test_must_method(self):
        q = (
            Query()
            .bool(filter=Term(field="field_a", value=2), _name="root_bool")
            .must({"exists": {"field": "field_b"}}, _name="root_bool")
            .must(Prefix(field="field_c", value="pre"), _name="root_bool")
            .must(
                Bool(
                    _name="level_1_bool",
                    should=[Range(field="field_d", gte=3), Ids(values=[1, 2, 3])],
                ),
                _name="root_bool",
            )
        )

        self.assertEqual(
            q.__str__(),
            """<Query>
bool
├── filter
│   └── term, field=field_a, value=2
└── must
    ├── exists, field=field_b
    ├── bool
    │   └── should
    │       ├── ids, values=[1, 2, 3]
    │       └── range, field=field_d, gte=3
    └── prefix, field=field_c, value="pre"
""",
        )

    def test_nested(self):
        q = Query()
        q1 = q.nested(
            query=Term(field="some_nested_field.other", value=2),
            path="some_nested_path",
            _name="nested_id",
        )
        self.assertEqual(
            q1.to_dict(),
            {
                "nested": {
                    "_name": "nested_id",
                    "path": "some_nested_path",
                    "query": {"term": {"some_nested_field.other": {"value": 2}}},
                }
            },
        )
        self.assertEqual(q1.root, "nested_id")

    def test_should(self):
        q = Query()
        q1 = q.should(
            Term(field="some_field", value=2),
            Term(field="other_field", value=3),
            _name="created_bool",
        )
        self.assertTrue(
            equal_queries(
                q1.to_dict(),
                {
                    "bool": {
                        "_name": "created_bool",
                        "should": [
                            {"term": {"some_field": {"value": 2}}},
                            {"term": {"other_field": {"value": 3}}},
                        ],
                    }
                },
            )
        )
        self.assertEqual(q1.root, "created_bool")

    def test_filter(self):
        q = Query()
        q1 = q.filter(
            Term(field="some_field", value=2),
            Term(field="other_field", value=3),
            _name="created_bool",
        )
        self.assertQueryEqual(
            q1.to_dict(),
            {
                "bool": {
                    "_name": "created_bool",
                    "filter": [
                        {"term": {"some_field": {"value": 2}}},
                        {"term": {"other_field": {"value": 3}}},
                    ],
                }
            },
        )
        self.assertEqual(q1.root, "created_bool")

    def test_must_not(self):
        q = Query()
        q1 = q.must_not(
            Term(field="some_field", value=2),
            Term(field="other_field", value=3),
            _name="created_bool",
        )
        self.assertTrue(
            equal_queries(
                q1.to_dict(),
                {
                    "bool": {
                        "_name": "created_bool",
                        "must_not": [
                            {"term": {"some_field": {"value": 2}}},
                            {"term": {"other_field": {"value": 3}}},
                        ],
                    }
                },
            )
        )
        self.assertEqual(q1.root, "created_bool")

    def test_deserialize_dict_query(self):
        # simple leaf
        q = Query({"term": {"some_field": {"value": 2}}})
        self.assertEqual(q.to_dict(), {"term": {"some_field": {"value": 2}}})
        self.assertEqual(len(q.list()), 1)
        n = q.get(q.root)
        self.assertIsInstance(n, TermNode)
        self.assertEqual(n.field, "some_field")

        # bool simple leaf
        q = Query({"bool": {"must_not": {"term": {"some_field": {"value": 2}}}}})
        self.assertEqual(len(q.list()), 3)
        n = q.get(q.root)
        self.assertIsInstance(n, BoolNode)
        self.assertEqual(
            q.to_dict(),
            {"bool": {"must_not": [{"term": {"some_field": {"value": 2}}}]}},
        )

        # bool multiple leaves
        q = Query(
            {
                "bool": {
                    "must_not": [
                        {"term": {"some_field": {"value": 2}}},
                        {"term": {"other_field": {"value": 3}}},
                    ]
                }
            }
        )
        self.assertEqual(len(q.list()), 4)
        n = q.get(q.root)
        self.assertIsInstance(n, BoolNode)
        self.assertQueryEqual(
            q.to_dict(),
            {
                "bool": {
                    "must_not": [
                        {"term": {"some_field": {"value": 2}}},
                        {"term": {"other_field": {"value": 3}}},
                    ]
                }
            },
        )

        # nested compound queries
        q = Query(
            {
                "nested": {
                    "path": "some_nested_path",
                    "query": {
                        "bool": {
                            "must_not": [
                                {"term": {"some_field": {"value": 2}}},
                                {"term": {"other_field": {"value": 3}}},
                            ]
                        }
                    },
                }
            }
        )
        self.assertEqual(len(q.list()), 6)
        n = q.get(q.root)
        self.assertIsInstance(n, NestedNode)
        self.assertQueryEqual(
            q.to_dict(),
            {
                "nested": {
                    "path": "some_nested_path",
                    "query": {
                        "bool": {
                            "must_not": [
                                {"term": {"some_field": {"value": 2}}},
                                {"term": {"other_field": {"value": 3}}},
                            ]
                        }
                    },
                }
            },
        )

    def test_query_method(self):
        # on empty query
        q = Query()
        q1 = q.query(Bool(_name="root_bool", must=Term(field="some_field", value=2)))
        q2 = q.query(
            {"bool": {"_name": "root_bool", "must": {"term": {"some_field": 2}}}}
        )
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
bool
└── must
    └── term, field=some_field, value=2
""",
            )

        # query WITHOUT defined parent -> must on top (existing bool)
        q = Query(Bool(_name="root_bool", must=Term(field="some_field", value=2)))
        q1 = q.query(Term(field="other_field", value=3))
        q2 = q.query({"term": {"other_field": 3}})
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
bool
└── must
    ├── term, field=some_field, value=2
    └── term, field=other_field, value=3
""",
            )

        # query WITHOUT defined parent -> must on top (non-existing bool)
        q = Query(Term(field="some_field", value=2))
        q1 = q.query(Term(field="other_field", value=3))
        q2 = q.query({"term": {"other_field": 3}})
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
bool
└── must
    ├── term, field=some_field, value=2
    └── term, field=other_field, value=3
""",
            )

        # query WITH defined parent (use default operator)
        q = Query(Nested(_name="root_nested", path="some_nested_path"))
        q1 = q.query(Term(field="some_nested_path.id", value=2), parent="root_nested")
        q2 = q.query({"term": {"some_nested_path.id": 2}}, parent="root_nested")
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
nested, path="some_nested_path"
└── query
    └── term, field=some_nested_path.id, value=2
""",
            )

        q = Query(Bool(_name="root_bool", filter=Term(field="some_field", value=2)))
        q1 = q.query(Term(field="some_other_field", value=2), parent="root_bool")
        q2 = q.query({"term": {"some_other_field": 2}}, parent="root_bool")
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
bool
├── filter
│   └── term, field=some_field, value=2
└── must
    └── term, field=some_other_field, value=2
""",
            )

        # query WITH defined parent and explicit operator
        q = Query(Bool(_name="root_bool", filter=Term(field="some_field", value=2)))
        q1 = q.query(
            Range(field="some_other_field", gte=3),
            parent="root_bool",
            parent_param="must_not",
        )
        q2 = q.query(
            {"range": {"some_other_field": {"gte": 3}}},
            parent="root_bool",
            parent_param="must_not",
        )
        for r in (q1, q2):
            self.assertEqual(
                r.__str__(),
                """<Query>
bool
├── filter
│   └── term, field=some_field, value=2
└── must_not
    └── range, field=some_other_field, gte=3
""",
            )

        # INVALID
        # query with leaf parent
        q = Query(Term(field="some_field", value=2, _name="leaf_node"))
        with self.assertRaises(ValueError) as e:
            q.query(Range(field="some_other_field", gte=3), parent="leaf_node")
        self.assertEqual(
            e.exception.args[0],
            "Cannot place clause under non-compound clause <leaf_node> of type <term>.",
        )
        with self.assertRaises(ValueError) as e:
            q.query({"range": {"some_other_field": {"gte": 3}}}, parent="leaf_node")
        self.assertEqual(
            e.exception.args[0],
            "Cannot place clause under non-compound clause <leaf_node> of type <term>.",
        )

        # query WITH wrong parent operator
        q = Query(Bool(_name="root_bool", filter=Term(field="some_field", value=2)))
        with self.assertRaises(ValueError) as e:
            q.query(
                Range(field="some_other_field", gte=3),
                parent="root_bool",
                parent_param="invalid_param",
            )
        self.assertEqual(
            e.exception.args[0],
            "Child operator <invalid_param> not permitted for compound query of type <Bool>",
        )
        with self.assertRaises(ValueError) as e:
            q.query(
                {"range": {"some_other_field": {"gte": 3}}},
                parent="root_bool",
                parent_param="invalid_param",
            )
        self.assertEqual(
            e.exception.args[0],
            "Child operator <invalid_param> not permitted for compound query of type <Bool>",
        )

        # query WITH non-existing parent
        q = Query(Bool(_name="root_bool", filter=Term(field="some_field", value=2)))
        with self.assertRaises(ValueError) as e:
            q.query(Range(field="some_other_field", gte=3), parent="yolo_id")
        self.assertEqual(
            e.exception.args[0], "Parent <yolo_id> does not exist in current query."
        )
        with self.assertRaises(ValueError) as e:
            q.query({"range": {"some_other_field": {"gte": 3}}}, parent="yolo_id")
        self.assertEqual(
            e.exception.args[0], "Parent <yolo_id> does not exist in current query."
        )

    def test_nested_query(self):
        q = (
            Query()
            .nested(path="some_nested", _name="nested_id")
            .query(Term(field="some_nested.type", value=2), parent="nested_id")
        )

        self.assertEqual(
            q.to_dict(),
            {
                "nested": {
                    "_name": "nested_id",
                    "path": "some_nested",
                    "query": {"term": {"some_nested.type": {"value": 2}}},
                }
            },
        )

    def test_must_below_nested_query(self):
        q = (
            Query()
            .nested(path="some_nested", _name="nested_id")
            .query(Term(field="some_nested.type", value=2), parent="nested_id")
            .query(
                Range(field="some_nested.creationYear", gte="2020"), parent="nested_id"
            )
        )

        self.assertTrue(
            equal_queries(
                q.to_dict(),
                {
                    "nested": {
                        "_name": "nested_id",
                        "path": "some_nested",
                        "query": {
                            "bool": {
                                "must": [
                                    {
                                        "range": {
                                            "some_nested.creationYear": {"gte": "2020"}
                                        }
                                    },
                                    {"term": {"some_nested.type": {"value": 2}}},
                                ]
                            }
                        },
                    }
                },
            )
        )

    def test_multiple_compound_on_top(self):
        q = (
            Query()
            .nested(path="some_nested_path", query=Term("some_nested_path.id", value=3))
            .should([Term(field="type", value=2)], _name="top")
        )
        self.assertEqual(
            q.to_dict(),
            {
                "bool": {
                    "_name": "top",
                    "must": [
                        {
                            "nested": {
                                "path": "some_nested_path",
                                "query": {
                                    "term": {"some_nested_path.id": {"value": 3}}
                                },
                            }
                        }
                    ],
                    "should": [{"term": {"type": {"value": 2}}}],
                }
            },
        )

    def test_multiple_must_below_nested_query(self):
        q1 = (
            Query()
            .query(Nested(path="some_nested", _name="nested_id"))
            .query(Range(field="some_nested.price", lte=100), parent="nested_id")
            .query(
                Range(field="some_nested.creationYear", gte="2020"), parent="nested_id"
            )
            .query(Term(field="some_nested.type", value=2), parent="nested_id")
        )

        q2 = (
            Query()
            .nested(path="some_nested", _name="nested_id")
            .query(Range(field="some_nested.price", lte=100), parent="nested_id")
            .query(
                Range(field="some_nested.creationYear", gte="2020"), parent="nested_id"
            )
            .query(Term(field="some_nested.type", value=2), parent="nested_id")
        )

        q3 = (
            Query()
            .nested(path="some_nested", _name="nested_id")
            .bool(
                parent="nested_id",
                must=[
                    Range(field="some_nested.creationYear", gte="2020"),
                    Range(field="some_nested.price", lte=100),
                    Term(field="some_nested.type", value=2),
                ],
            )
        )

        q4 = (
            Query()
            .query({"nested": {"path": "some_nested", "_name": "nested_id"}})
            .query(
                {
                    "bool": {
                        "must": [
                            {"range": {"some_nested.creationYear": {"gte": "2020"}}},
                            {"range": {"some_nested.price": {"lte": 100}}},
                            {"term": {"some_nested.type": {"value": 2}}},
                        ]
                    }
                },
                parent="nested_id",
            )
        )

        expected = {
            "nested": {
                "_name": "nested_id",
                "path": "some_nested",
                "query": {
                    "bool": {
                        "must": [
                            {"range": {"some_nested.creationYear": {"gte": "2020"}}},
                            {"range": {"some_nested.price": {"lte": 100}}},
                            {"term": {"some_nested.type": {"value": 2}}},
                        ]
                    }
                },
            }
        }
        for i, q in enumerate((q1, q2, q3, q4)):
            self.assertTrue(
                equal_queries(q.to_dict(), expected), "failed on %d" % (i + 1)
            )

        q5 = (
            Query()
            .nested(path="some_nested", _name="nested_id")
            .must(
                Range(field="some_nested.creationYear", gte="2020"),
                parent="nested_id",
                _name="bool_id",
            )
            .must(Range(field="some_nested.price", lte=100), _name="bool_id")
            .must(Term(field="some_nested.type", value=2), _name="bool_id")
        )
        self.assertEqual(
            ordered(q5.to_dict()),
            ordered(
                {
                    "nested": {
                        "_name": "nested_id",
                        "path": "some_nested",
                        "query": {
                            "bool": {
                                "_name": "bool_id",
                                "must": [
                                    {
                                        "range": {
                                            "some_nested.creationYear": {"gte": "2020"}
                                        }
                                    },
                                    {"term": {"some_nested.type": {"value": 2}}},
                                    {"range": {"some_nested.price": {"lte": 100}}},
                                ],
                            }
                        },
                    }
                }
            ),
        )

    def test_autonested(self):
        q = Query(
            mapping={
                "properties": {
                    "actors": {
                        "type": "nested",
                        "properties": {"id": {"type": "keyword"}},
                    }
                }
            },
            nested_autocorrect=True,
        )
        self.assertEqual(
            q.query("term", actors__id=2).to_dict(),
            {
                "nested": {
                    "path": "actors",
                    "query": {"term": {"actors.id": {"value": 2}}},
                }
            },
        )

    def test_query_unnamed_inserts(self):

        q = (
            Query()
            .must(Terms(genres=["Action", "Thriller"]))
            .must(Range("rank", gte=7))
            .query(Nested(_name="nested_roles", path="roles"))
        )

        # we name the nested query that we would potentially use
        # but a compound clause (bool, nested etc..) without any children clauses is not serialized
        self.assertEqual(
            ordered(q.to_dict()),
            ordered(
                {
                    "bool": {
                        "must": [
                            {"terms": {"genres": ["Action", "Thriller"]}},
                            {"range": {"rank": {"gte": 7}}},
                        ]
                    }
                }
            ),
        )

        # we declare that those clauses must be placed below 'nested_roles' condition
        q = q.query(Term("roles.gender", value="F"), parent="nested_roles").query(
            Term("roles.role", value="Reporter"), parent="nested_roles"
        )

        self.assertTrue(
            equal_queries(
                q.to_dict(),
                {
                    "bool": {
                        "must": [
                            {"terms": {"genres": ["Action", "Thriller"]}},
                            {"range": {"rank": {"gte": 7}}},
                            {
                                "nested": {
                                    "_name": "nested_roles",
                                    "path": "roles",
                                    "query": {
                                        "bool": {
                                            "must": [
                                                {
                                                    "term": {
                                                        "roles.gender": {"value": "F"}
                                                    }
                                                },
                                                {
                                                    "term": {
                                                        "roles.role": {
                                                            "value": "Reporter"
                                                        }
                                                    }
                                                },
                                            ]
                                        }
                                    },
                                }
                            },
                        ]
                    }
                },
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .aggs import AbstractParentAgg


class Global(AbstractParentAgg):
    KEY = "global"


class Filter(AbstractParentAgg):
    KEY = "filter"


class Nested(AbstractParentAgg):
    KEY = "nested"


class ReverseNested(AbstractParentAgg):
    KEY = "reverse_nested"


class Missing(AbstractParentAgg):
    KEY = "missing"


class Terms(AbstractParentAgg):
    KEY = "terms"


class Filters(AbstractParentAgg):
    KEY = "filters"


class Histogram(AbstractParentAgg):
    KEY = "histogram"


class DateHistogram(AbstractParentAgg):
    KEY = "date_histogram"


class Range(AbstractParentAgg):
    KEY = "range"


class DateRange(AbstractParentAgg):
    KEY = "date_range"


class Composite(AbstractParentAgg):
    KEY = "composite"



#!/usr/bin/env python
# coding: utf-8

# # IMDB exploration with Pandagg

# This tutorial will guide you in some of pandagg functionalities, exploring IMDB data.
#
# 1. Cluster indices discovery
# 2. Mapping exploration
# 3. Aggregations
# 4. Queries
#

# In[1]:


# requires to be declared on top
import matplotlib.pyplot as plt
import seaborn


# ## 1. Cluster indices discovery

# In[2]:


# instanciate client just as you would do with regular elastic client
from pandagg import Elasticsearch

client = Elasticsearch(hosts=["localhost:9300"])


# In[3]:


# indices instance lists available indices of cluster, with their mappings and settings
indices = client.fetch_indices()
indices


# Indices are accessible with autocompletion
#
# ![autocomplete](ressources/autocomplete_index.png)

# In[4]:


movies = indices.movies
movies


# ## 2. Mapping exploration

# In[5]:


# mapping is accessible via "mapping" attribute of "Index" instance
m = movies.mapping

# this is equivalent to instanciate Mapping from dict as follow:
from examples.imdb.load import mapping as imdb_mapping
from pandagg.mapping import IMapping

m = IMapping(imdb_mapping, client=client, index_name="movies")

# Note: client and index_name arguments are optional, but doing so provides the ability to
# compute "quick-access" aggregations on fields (will be detailed below)


# In[6]:


# you can navigate (with help of tab autocomplete) into your mapping
m.directors


# In[7]:


# going deeper
m.directors.full_name


# In[8]:


# calling instance will display mapping definition
m.directors.full_name()


# ### Quick access aggregations from mapping

# Mapping leaves (here genres) all have a "a" attribute (for aggregation).
# Autocomplete will display all possible aggregations types on this field
#
# ![autocomplete](ressources/autocomplete_agg.png)
#

# In[9]:


# passed parameters will be added to aggregation body
m.genres.a.terms(missing="N/A", size=5)


# In[10]:


m.rank.a.stats()


# In[11]:


# query parameter enable to filter on some conditions
# documentaries are overall better ranked than average of movies
m.rank.a.stats(query={"term": {"genres": "Documentary"}})


# ## 3. Aggregations

# Let's compute the number of movies per decade per genre.
#

# ### Regular declaration

# In[12]:


regular_syntax = {
    "genres": {
        "terms": {"field": "genres", "size": 3},
        "aggs": {
            "movie_decade": {
                "date_histogram": {"field": "year", "fixed_interval": "3650d"}
            }
        },
    }
}

from pandagg.aggs import Aggs

agg = Aggs(regular_syntax)

assert agg.to_dict() == regular_syntax

agg


# ### DSL syntax
# The following syntaxes are strictly equivalent to the above one:

# In[13]:


from pandagg.aggs import DateHistogram, Terms

agg_dsl = Aggs(
    Terms(
        "genres",
        field="genres",
        size=3,
        aggs=DateHistogram(name="movie_decade", field="year", fixed_interval="3650d"),
    )
)

# or using groupby method: the listed aggregations will be placed from top to bottom:

agg_variant = Aggs().groupby(
    [
        Terms("genres", field="genres", size=3),
        DateHistogram("movie_decade", field="year", fixed_interval="3650d"),
    ]
)


assert agg_dsl.to_dict() == agg_variant.to_dict()
assert agg_dsl.to_dict() == regular_syntax

# decade = DateHistogram('movie_decade', field='year', fixed_interval='3650d')
# per_decate_genres = movies.groupby(['genres', decade],size=3).execute()
agg_dsl


# #### About groupby and agg methods
#
# - `groupby` method will arrange passed aggregations clauses "vertically" (nested manner),
# - `agg` method will arrange them "horizontally"

# In[14]:


Aggs().groupby(
    [
        Terms("genres", field="genres", size=3),
        DateHistogram("movie_decade", field="year", fixed_interval="3650d"),
    ]
)


# In[15]:


Aggs().aggs(
    [
        Terms("genres", field="genres", size=3),
        DateHistogram("movie_decade", field="year", fixed_interval="3650d"),
    ]
)


# Both `groupby` and `agg` will place provided aggregations under the `insert_below` (parent id) aggregation clause if `insert_below` is provided, else under the deepest bucket aggregation if there is no ambiguity:
# ```
# OK: A──> B ─> C ─> NEW_AGGS
#
# KO: A──> B
#     └──> C
# ```

# In[16]:


# taking again this example
example_agg = Aggs(regular_syntax)
example_agg


# In[17]:


# groupby behaviour
example_agg.groupby(["roles.role", "roles.gender"], insert_below="genres")


# In[18]:


# agg behaviour
example_agg.aggs(["roles.role", "roles.gender"], insert_below="genres")


# ### Aggregation execution and parsing

# Aggregation instance can be bound to an Elasticsearch client, either at `__init__`, either using `bind` method.

# In[19]:


agg_dsl.bind(client=client, index_name="movies")


# Doing so provides the ability to execute aggregation request, and parse the response in multiple formats. Formats will be detailed in next example, here we use the dataframe format:
#
# *Note: requires to install **pandas** dependency*

# In[20]:


per_decate_genres = agg_dsl.execute(output="dataframe")
per_decate_genres.unstack()


# In[21]:


per_decate_genres.unstack().T.plot(figsize=(12, 12))


# **Another example:**
# who are the actors who have played in the highest number of movies between 1990 and 2000, and what was the average ranking of the movies they played in per genre?
#

# In[22]:


from datetime import datetime
from pandagg.aggs import Aggs, Avg, Min, Max
from pandagg.query import Range


# in groupby and agg methods,
agg = (
    Aggs(client=client, index_name="movies", mapping=imdb_mapping)
    .groupby(["roles.full_name.raw", "genres"], size=2)
    .aggs(
        [
            Avg("avg_rank", field="rank"),
            Min("min_date", field="year"),
            Max("max_date", field="year"),
        ]
    )
    .query(Range(field="year", gte="2000", lt="2010"))
)

print(agg)
r = agg.execute()

r["min_year"] = r.min_date.apply(lambda x: datetime.fromtimestamp(x / 1000.0).year)
r["max_year"] = r.max_date.apply(lambda x: datetime.fromtimestamp(x / 1000.0).year)
r


# #### As raw output

# In[23]:


# agg.execute(output='raw')


# #### As interactive tree

# In[24]:


t = agg.execute(output="tree")
t


# #### Navigation with autocompletion

# In[25]:


t.roles_full_name_raw_Grey_DeLisle__599599_


# #### List documents in given bucket (with autocompletion)

# In[26]:


delisle_adventure = t.roles_full_name_raw_Grey_DeLisle__599599_.reverse_nested_below_roles_full_name_raw.genres_Adventure.list_documents(
    _source=["id", "genres", "name"], size=2
)


# In[27]:


delisle_adventure


# ## 4. Queries

# Suppose I want:
# - actions or thriller movies
# - with ranking >= 7
# - with a female actor playing a reporter role

# ### Regular syntax

# We would perform the following request:

# In[28]:


expected_query = {
    "bool": {
        "must": [
            {"terms": {"genres": ["Action", "Thriller"]}},
            {"range": {"rank": {"gte": 7}}},
            {
                "nested": {
                    "path": "roles",
                    "query": {
                        "bool": {
                            "must": [
                                {"term": {"roles.gender": {"value": "F"}}},
                                {"term": {"roles.role": {"value": "Reporter"}}},
                            ]
                        }
                    },
                }
            },
        ]
    }
}


# We can build our Query instance using this regular syntax:

# In[29]:


from pandagg.query import Query

q = Query(expected_query)
q


# ### DSL syntax

# With pandagg DSL syntax, it could also be declared this way:

# In[30]:


from pandagg.query import Nested, Bool, Query, Range, Term, Terms as TermsFilter

# warning, pandagg.query.Terms and pandagg.agg.Terms classes have same name, but one is a filter, the other an aggreggation

q = Query(
    Bool(
        must=[
            TermsFilter("genres", terms=["Action", "Thriller"]),
            Range("rank", gte=7),
            Nested(
                path="roles",
                query=Bool(
                    must=[
                        Term("roles.gender", value="F"),
                        Term("roles.role", value="Reporter"),
                    ]
                ),
            ),
        ]
    )
)


# In[31]:


# query computation
q.to_dict() == expected_query


# Suppose you want to expose a route to your customers with actionable filters, it is easy to add query clauses at specific places in your query by chaining your clauses:
#

# In[32]:


# accepts mix of DSL and dict syntax

my_query = (
    Query()
    .must(TermsFilter("genres", terms=["Action", "Thriller"]))
    .must({"range": {"rank": {"gte": 7}}})
    .must(
        Nested(
            path="roles",
            query=Bool(
                must=[
                    {"term": {"roles.gender": {"value": "F"}}},
                    {"term": {"roles.role": {"value": "Reporter"}}},
                ]
            ),
        )
    )
)

my_query.to_dict() == expected_query


# ### Advanced query declaration using _named queries_
#
# We can take advantage of [named queries](https://www.elastic.co/guide/en/elasticsearch/reference/6.8/search-request-named-queries-and-filters.html) to specifically declare where we want to insert a clause.
#
# A simple use case could be to expose some filters to a client among which some apply to nested clauses (for instance nested 'roles').

# In[33]:


from pandagg.utils import equal_queries

# suppose API exposes those filters
genres_in = ["Action", "Thriller"]
rank_above = 7
filter_role_gender = "F"
filter_role = "Reporter"


q = Query()


if genres_in is not None:
    q = q.must(TermsFilter("genres", terms=genres_in))
if rank_above is not None:
    q = q.must(Range("rank", gte=rank_above))

# we name the nested query that we would potentially use
q = q.query(Nested(_name="nested_roles", path="roles"))
# a compound clause (bool, nested etc..) without any children clauses is not serialized
assert q.to_dict() == {
    "bool": {
        "must": [
            {"terms": {"genres": ["Action", "Thriller"]}},
            {"range": {"rank": {"gte": 7}}},
        ]
    }
}


# we declare that those clauses must be placed below 'nested_roles' condition
if filter_role_gender is not None:
    q = q.query(Term("roles.gender", value=filter_role_gender), parent="nested_roles")
if filter_role is not None:
    q = q.query(Term("roles.role", value=filter_role), parent="nested_roles")

assert equal_queries(q.to_dict(), expected_query)
q


# In[ ]:



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from .aggs import AbstractLeafAgg


class TopHits(AbstractLeafAgg):
    KEY = "top_hits"


class Avg(AbstractLeafAgg):
    KEY = "avg"


class Sum(AbstractLeafAgg):
    KEY = "sum"


class Max(AbstractLeafAgg):
    KEY = "max"


class Min(AbstractLeafAgg):
    KEY = "min"


class Cardinality(AbstractLeafAgg):
    KEY = "cardinality"


class Stats(AbstractLeafAgg):
    KEY = "stats"


class ExtendedStats(AbstractLeafAgg):
    KEY = "extended_stats"


class GeoBound(AbstractLeafAgg):
    KEY = "geo_bounds"


class GeoCentroid(AbstractLeafAgg):
    KEY = "geo_centroid"


class Percentiles(AbstractLeafAgg):
    """Percents body argument can be passed to specify which percentiles to fetch."""

    KEY = "percentiles"


class PercentileRanks(AbstractLeafAgg):
    KEY = "percentile_ranks"


class ValueCount(AbstractLeafAgg):
    KEY = "value_count"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from pandagg.node._node import Node


class Bucket(Node):

    ROOT_NAME = "root"

    def __init__(self, value, key=None, level=None):
        self.value = value
        self.level = level if level is not None else self.ROOT_NAME
        self.key = key
        super(Bucket, self).__init__()

    @property
    def attr_name(self):
        """Determine under which attribute name the bucket will be available in response tree.
        Dots are replaced by `_` characters so that they don't prevent from accessing as attribute.

        Resulting attribute unfit for python attribute name syntax is still possible and will be accessible through
        item access (dict like), see more in 'utils.Obj' for more details.
        """
        if self.key is not None:
            return "%s_%s" % (self.level.replace(".", "_"), self.key)
        return self.level.replace(".", "_")

    def line_repr(self, **kwargs):
        REPR_SIZE = 60
        s = self.level
        if self.key is not None:
            s += "=%s" % self.key
        if self.value is not None:
            pad = max(
                REPR_SIZE - 4 * kwargs.get("depth") - len(s) - len(str(self.value)), 4
            )
            s = s + " " * pad + str(self.value)
        return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from builtins import str as text
import json

from pandagg.node._node import Node


class QueryClause(Node):
    _type_name = "query"
    KEY = None

    def __init__(self, **body):
        body = body.copy()
        _name = body.pop("_name", None)
        self.body = body
        self._named = _name is not None
        super(QueryClause, self).__init__(identifier=_name)

    def line_repr(self, depth, **kwargs):
        if not self.body:
            return self.KEY
        return ", ".join([text(self.KEY), self._params_repr(self.body)])

    @staticmethod
    def _params_repr(params):
        params = params or {}
        return ", ".join(
            "%s=%s" % (text(k), text(json.dumps(params[k], sort_keys=True)))
            for k in sorted(params.keys())
        )

    @property
    def name(self):
        return self.identifier

    @property
    def _identifier_prefix(self):
        return "%s_" % self.KEY

    def to_dict(self, with_name=True):
        b = self.body.copy()
        if with_name and self._named:
            b["_name"] = self.name
        return {self.KEY: b}

    def __str__(self):
        return "<{class_}, id={id}, type={type}, body={body}>".format(
            class_=text(self.__class__.__name__),
            type=text(self.KEY),
            id=text(self.identifier),
            body=json.dumps(self.body),
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return other.to_dict() == self.to_dict()
        # make sure we still equal to a dict with the same data
        return other == self.to_dict()


class LeafQueryClause(QueryClause):
    pass


class AbstractSingleFieldQueryClause(LeafQueryClause):
    _FIELD_AT_BODY_ROOT = False

    def __init__(self, field, _name=None, **body):
        self.field = field
        if self._FIELD_AT_BODY_ROOT:
            super(LeafQueryClause, self).__init__(_name=_name, field=field, **body)
        else:
            super(LeafQueryClause, self).__init__(_name=_name, **body)


class FlatFieldQueryClause(AbstractSingleFieldQueryClause):
    """Query clause applied on one single field.
    Example:

    Exists:
    {"exists": {"field": "user"}}
    -> field = "user"
    -> body = {"field": "user"}
    q = Exists(field="user")

    DistanceFeature:
    {"distance_feature": {"field": "production_date", "pivot": "7d", "origin": "now"}}
    -> field = "production_date"
    -> body = {"field": "production_date", "pivot": "7d", "origin": "now"}
    q = DistanceFeature(field="production_date", pivot="7d", origin="now")
    """

    _FIELD_AT_BODY_ROOT = True

    def __init__(self, field, _name=None, **body):
        self.field = field
        super(FlatFieldQueryClause, self).__init__(_name=_name, field=field, **body)


class KeyFieldQueryClause(AbstractSingleFieldQueryClause):
    """Clause with field used as key in clause body:

    Term:
    {"term": {"user": {"value": "Kimchy", "boost": 1}}}
    -> field = "user"
    -> body = {"user": {"value": "Kimchy", "boost": 1}}
    q1 = Term(user={"value": "Kimchy", "boost": 1}})
    q2 = Term(field="user", value="Kimchy", boost=1}})

    Can accept a "_implicit_param" attribute specifying which is the equivalent key when inner body isn't a dict but a
    raw value.
    For Term:
    _implicit_param = "value"
    q = Term(user="Kimchy")
    {"term": {"user": {"value": "Kimchy"}}}
    -> field = "user"
    -> body = {"term": {"user": {"value": "Kimchy"}}}
    """

    _implicit_param = None

    def __init__(self, field=None, _name=None, _expand__to_dot=True, **params):
        if field is None:
            # Term(item__id=32) or Term(item__id={'value': 32, 'boost': 1})
            if len(params) != 1:
                raise ValueError(
                    "Invalid declaration for <%s> clause, got:\n%s"
                    % (self.__class__.__name__, params)
                )
            if _expand__to_dot:
                field, value = self.expand__to_dot(params).copy().popitem()
            else:
                field, value = params.copy().popitem()
            params = value if isinstance(value, dict) else {self._implicit_param: value}
        self.inner_body = params
        super(KeyFieldQueryClause, self).__init__(
            field=field, _name=_name, **{field: params}
        )

    def line_repr(self, depth, **kwargs):
        if not self.inner_body:
            return ", ".join([text(self.KEY), "field=%s" % text(self.field)])
        return ", ".join(
            [
                text(self.KEY),
                "field=%s" % text(self.field),
                self._params_repr(self.inner_body),
            ]
        )


class MultiFieldsQueryClause(LeafQueryClause):
    def __init__(self, fields, _name=None, **body):
        self.fields = fields
        super(LeafQueryClause, self).__init__(_name=_name, fields=fields, **body)

    def line_repr(self, depth, **kwargs):
        return "%s, fields=%s" % (self.KEY, list(map(text, self.fields)))


class ParentParameterClause(QueryClause):
    MULTIPLE = False



from pandagg.node.query.abstract import QueryClause


class CompoundClause(QueryClause):
    """Compound clauses can encapsulate other query clauses::

        {
            "<query_type>" : {
                <query_body>
                <children_clauses>
            }
        }
    """

    _default_operator = None
    _parent_params = None

    @classmethod
    def operator(cls, key):
        if key is None:
            return cls.get_dsl_class(cls._default_operator)
        if key not in cls._parent_params:
            raise ValueError(
                "Child operator <%s> not permitted for compound query of type <%s>"
                % (key, cls.__name__)
            )
        return cls.get_dsl_class(key)


class Bool(CompoundClause):
    _default_operator = "must"
    _parent_params = ["should", "must", "must_not", "filter"]
    KEY = "bool"


class Boosting(CompoundClause):
    _default_operator = "positive"
    _parent_params = ["positive", "negative"]
    KEY = "boosting"


class ConstantScore(CompoundClause):
    _default_operator = "filter"
    _parent_params = ["filter", "boost"]
    KEY = "constant_score"


class DisMax(CompoundClause):
    _default_operator = "queries"
    _parent_params = ["queries"]
    KEY = "dis_max"


class FunctionScore(CompoundClause):
    _default_operator = "query"
    _parent_params = ["query"]
    KEY = "function_score"



from .abstract import KeyFieldQueryClause, AbstractSingleFieldQueryClause


class GeoBoundingBox(KeyFieldQueryClause):
    KEY = "geo_bounding_box"


class GeoDistance(AbstractSingleFieldQueryClause):
    KEY = "geo_distance"

    def __init__(self, distance, **body):
        # pop all allowed args to find out which keyword is used as field
        _name = body.pop("_name", None)
        distance_type = body.pop("distance_type", None)
        validation_method = body.pop("validation_method", None)
        if len(body) != 1:
            raise ValueError("Wrong declaration: %s" % body)
        field, location = self.expand__to_dot(body).popitem()
        self.field = field
        b = {field: location}
        if distance_type is not None:
            b["distance_type"] = distance_type
        if validation_method is not None:
            b["validation_method"] = validation_method
        super(GeoDistance, self).__init__(
            _name=_name, field=field, distance=distance, **b
        )

    def line_repr(self, depth, **kwargs):
        return "%s, field=%s" % (self.KEY, self.field)


class GeoPolygone(KeyFieldQueryClause):
    KEY = "geo_polygon"


class GeoShape(KeyFieldQueryClause):
    KEY = "geo_shape"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from future.utils import python_2_unicode_compatible

from builtins import str as text

from lighttree import Tree as OriginalTree
from six import add_metaclass

from pandagg.utils import DslMeta, get_dsl_class


@python_2_unicode_compatible
@add_metaclass(DslMeta)
class Tree(OriginalTree):

    KEY = None
    _type_name = None

    get_dsl_class = classmethod(get_dsl_class)

    @classmethod
    def get_node_dsl_class(cls, name):
        return cls.node_class.get_dsl_class(name)

    def __str__(self):
        return "<{class_}>\n{tree}".format(
            class_=text(self.__class__.__name__), tree=self.show(limit=40)
        )

    def __repr__(self):
        return self.__str__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from builtins import str as text

from future.utils import python_2_unicode_compatible, iteritems, string_types

from pandagg.node.query._parameter_clause import ParentParameterClause
from pandagg.node.query.abstract import QueryClause, LeafQueryClause
from pandagg.node.query.compound import CompoundClause, Bool as BoolNode
from pandagg.node.query.joining import Nested

from pandagg.tree._tree import Tree
from pandagg.tree.mapping import Mapping

ADD = "add"
REPLACE = "replace"
REPLACE_ALL = "replace_all"


@python_2_unicode_compatible
class Query(Tree):
    r"""Combination of query clauses.

    Mapping declaration is optional, but doing so validates query validity and automatically inserts nested clauses
    when necessary.

    :Keyword Arguments:
        * *mapping* (``dict`` or ``pandagg.tree.mapping.Mapping``) --
          Mapping of requested indice(s). Providing it will add validation features, and add required nested
          clauses if missing.

        * *nested_autocorrect* (``bool``) --
          In case of missing nested clauses in query, if True, automatically add missing nested clauses, else raise
          error.

        * remaining kwargs:
          Used as body in query clauses.
    """

    _type_name = "query_tree"
    KEY = None
    node_class = QueryClause

    def __init__(self, *args, **kwargs):
        self.mapping = Mapping(kwargs.pop("mapping", None))
        self.nested_autocorrect = kwargs.pop("nested_autocorrect", False)
        super(Query, self).__init__()
        if args or kwargs:
            self._fill(*args, **kwargs)

    def __nonzero__(self):
        return bool(self.to_dict())

    __bool__ = __nonzero__

    def _clone_init(self, deep=False):
        return Query(
            mapping=self.mapping.clone(with_tree=True, deep=deep),
            nested_autocorrect=self.nested_autocorrect,
        )

    @classmethod
    def _from_dict(cls, d):
        """return Query, from dict"""
        # {"nested": {"path": "xxx", "query": {"term": {"some_field": "234"}}}}
        if not len(d.keys()) == 1:
            raise ValueError("Wrong declaration, got %s" % d.keys())
        vk, vv = d.copy().popitem()
        return cls._get_dsl_class_from_tree_or_node(vk, **vv)

    @classmethod
    def _get_dsl_class_from_tree_or_node(cls, key, **body):
        # either a compound query clause -> search among trees
        if key in cls._classes:
            return cls.get_dsl_class(key)(**body)
        # either a simple query clause -> search among nodes
        elif key in cls.node_class._classes:
            return Query(cls.node_class.get_dsl_class(key)(**body))
        else:
            raise ValueError("Unkown clause %s" % key)

    def _fill(self, *args, **kwargs):
        if kwargs:
            # only allowed when using special syntax: Query("term", my__field=23)
            if len(args) != 1:
                raise ValueError("Unkown syntax %s %s" % (args, kwargs))
            arg = args[0]
            self.insert(self._get_dsl_class_from_tree_or_node(arg, **kwargs))
            return

        if not args:
            return

        if len(args) > 1:
            raise ValueError("Wrong declaration")

        arg = args[0]
        # None
        if arg is None:
            return self
        # Term(), Query
        if isinstance(arg, (Query, QueryClause)):
            return self.insert(arg)
        # {"bool": {"filter: xxx"}}
        if isinstance(arg, dict):
            self.insert(self._from_dict(arg))
            return
        raise ValueError("Unkown declaration %s" % arg)

    def _insert_node_below(self, node, parent_id=None):
        """Override lighttree.Tree._insert_node_below method to ensure inserted query clause is consistent."""
        if parent_id is not None:
            pnode = self.get(parent_id)
            if isinstance(pnode, LeafQueryClause):
                raise ValueError(
                    "Cannot add clause under leaf query clause <%s>" % pnode.KEY
                )
            if isinstance(pnode, ParentParameterClause):
                if isinstance(node, ParentParameterClause):
                    raise ValueError(
                        "Cannot add parameter clause <%s> under another parameter clause <%s>"
                        % (pnode.KEY, node.KEY)
                    )
            if isinstance(pnode, CompoundClause):
                if (
                    not isinstance(node, ParentParameterClause)
                    or node.KEY not in pnode._parent_params
                ):
                    raise ValueError(
                        "Expect a parameter clause of type %s under <%s> compound clause, got <%s>"
                        % (pnode._parent_params, pnode.KEY, node.KEY)
                    )

        # automatic handling of nested clauses
        if isinstance(node, Nested) or not self.mapping or not hasattr(node, "field"):
            return super(Query, self)._insert_node_below(node=node, parent_id=parent_id)
        required_nested_level = self.mapping.nested_at_field(node.field)
        if self.is_empty():
            current_nested_level = None
        else:
            current_nested_level = self.applied_nested_path_at_node(parent_id)
        if current_nested_level == required_nested_level:
            return super(Query, self)._insert_node_below(node=node, parent_id=parent_id)
        if not self.nested_autocorrect:
            raise ValueError(
                "Invalid %s query clause on %s field. Invalid nested: expected %s, current %s."
                % (node.KEY, node.field, required_nested_level, current_nested_level)
            )
        # requires nested - apply all required nested fields
        to_insert = node
        for nested_lvl in self.mapping.list_nesteds_at_field(node.field):
            if current_nested_level != nested_lvl:
                to_insert = self.get_dsl_class("nested")(
                    path=nested_lvl, query=to_insert
                )
        super(Query, self).insert(to_insert, parent_id)

    def applied_nested_path_at_node(self, nid):
        # from current node to root
        for id_ in [nid] + self.ancestors(nid):
            node = self.get(id_)
            if isinstance(node, Nested):
                return node.path
        return None

    def to_dict(self, from_=None, with_name=True):
        """Return None if no query clause.
        """
        if self.root is None:
            return None
        from_ = self.root if from_ is None else from_
        node = self.get(from_)
        if isinstance(node, LeafQueryClause):
            return node.to_dict(with_name=True)
        serialized_children = []
        should_yield = False
        for child_node in self.children(node.identifier, id_only=False):
            serialized_child = self.to_dict(
                from_=child_node.identifier, with_name=with_name
            )
            if serialized_child is not None:
                serialized_children.append(serialized_child)
                should_yield = True
        if not should_yield:
            return None
        if isinstance(node, CompoundClause):
            # {bool: {filter: ..., must: ...}
            q = node.to_dict(with_name=with_name)
            extra_body = {k: v for d in serialized_children for k, v in d.items()}
            q[node.KEY].update(extra_body)
            return q
        # parent parameter clause
        # {filter: [{...}, {...}]}
        assert isinstance(node, ParentParameterClause)
        if node.MULTIPLE:
            return {node.KEY: serialized_children}
        return {node.KEY: serialized_children[0]}

    def query(self, *args, **kwargs):
        mode = kwargs.pop("mode", ADD)
        parent = kwargs.pop("parent", None)
        parent_param = kwargs.pop("parent_param", None)
        child = kwargs.pop("child", None)
        child_param = kwargs.pop("child_param", None)
        return self._insert_into(
            Query(*args, **kwargs),
            parent=parent,
            child=child,
            mode=mode,
            child_param=child_param,
            parent_param=parent_param,
        )

    def _insert_into(
        self,
        inserted,
        mode,
        parent=None,
        parent_param=None,
        child=None,
        child_param=None,
    ):
        """Insert element (node or tree) in query.

        If compound query with existing identifier: merge according to mode (place in-between parent and child).
        If no parent nor child is provided, place on top (wrapped in bool-must if necessary).
        If a child is provided (only possible if inserted node is compound): place on top using child_param.
        If a parent is provided (only under compound query): place under it.
        """
        if not isinstance(inserted, Query):
            inserted = Query(inserted)
        if inserted.is_empty():
            return self
        # ensure we don't modify passed query
        inserted = inserted.clone(with_tree=True)
        inserted_root = inserted.get(inserted.root)
        q = self.clone(with_tree=True)

        # If compound query with existing name: merge according to mode (place in-between parent and child).
        if isinstance(inserted_root, CompoundClause) and inserted_root.name in q:
            if child is not None or parent is not None:
                raise ValueError(
                    "Child or parent cannot be provided when inserting compound clause with existing "
                    "_name <%s> in query. Got child <%s> and parent <%s>."
                    % (inserted_root.name, child, parent)
                )
            return q._compound_update(
                name=inserted.root, new_compound=inserted, mode=mode
            )

        # If no parent nor child is provided, place on top (wrapped in bool-must if necessary).
        if parent is None and child is None:
            # if inital query is empty, just insert new one
            if q.root is None:
                return q.insert(inserted)
            # if both initial root query and inserted one are bool, merge
            if isinstance(q.get(q.root), BoolNode) and isinstance(
                inserted_root, BoolNode
            ):
                return q._compound_update(name=q.root, new_compound=inserted, mode=mode)
            # if only inserted node is bool, insert initial query in it
            if isinstance(inserted_root, BoolNode):
                child_operator = inserted_root.operator(child_param)
                child_operator_node = next(
                    (
                        c
                        for c in inserted.children(inserted.root, id_only=False)
                        if isinstance(c, child_operator)
                    ),
                    None,
                )
                if child_operator_node is None:
                    child_operator_node = child_operator()
                    inserted.insert_node(child_operator_node, parent_id=inserted.root)
                return inserted.insert(item=q, parent_id=child_operator_node.name)
            if isinstance(q.get(q.root), BoolNode):
                return q.must(
                    inserted,
                    _name=q.root,
                    mode=mode,
                    parent_param=parent_param,
                    child_param=child_param,
                )
            parent_param_key = BoolNode.operator(parent_param).KEY
            return q.bool(
                parent_param=parent_param,
                child_param=child_param,
                mode=mode,
                **{parent_param_key: inserted}
            )

        # If a child is provided (only possible if inserted node is compound): place on top using child_param.
        if child is not None:
            if not isinstance(inserted_root, CompoundClause):
                raise ValueError(
                    "Cannot place non-compound clause <%s> above other clause <%s>."
                    % (inserted_root.KEY, child)
                )
            if child not in q:
                raise ValueError("Child <%s> does not exist in current query." % child)
            child_operator = inserted_root.operator(child_param)
            if parent is not None:
                raise ValueError(
                    "Cannot declare both parent <%s> and child <%s> (only one accepted)."
                    % (parent, child)
                )

            # suppose we are under a nested clause, the parent is the "query" param clause
            existing_parent_param_node = q.parent(child, id_only=False)
            direct_pid = (
                existing_parent_param_node.name if existing_parent_param_node else None
            )
            child_tree = q.drop_subtree(child)

            q.insert(inserted, parent_id=direct_pid)
            child_operator_node = next(
                (
                    c
                    for c in q.children(inserted.root, id_only=False)
                    if isinstance(c, child_operator)
                ),
                None,
            )
            if child_operator_node is None:
                child_operator_node = child_operator()
                q.insert_node(child_operator_node, parent_id=inserted.root)
            q.insert(item=child_tree, parent_id=child_operator_node.name)
            return q

        # If a parent is provided (only under compound query): place under it.
        if parent not in q:
            raise ValueError("Parent <%s> does not exist in current query." % parent)
        parent_node = q.get(parent)
        if not isinstance(parent_node, CompoundClause):
            raise ValueError(
                "Cannot place clause under non-compound clause <%s> of type <%s>."
                % (parent, parent_node.KEY)
            )
        parent_operator = parent_node.operator(parent_param)
        parent_operator_node = next(
            (
                c
                for c in q.children(parent, id_only=False)
                if isinstance(c, parent_operator)
            ),
            None,
        )
        if parent_operator_node is not None and not parent_operator_node.MULTIPLE:
            if isinstance(parent_node, BoolNode):
                return q.bool(must=inserted, _name=parent)
            child_node = q.children(parent_operator_node.name, id_only=False)[0]
            child = child_node.name
            if isinstance(child_node, BoolNode):
                return q.bool(must=inserted, _name=child, mode=mode)
            return q.bool(must=inserted, child=child, mode=mode)
        if parent_operator_node is None:
            parent_operator_node = parent_operator()
            q.insert_node(parent_operator_node, parent_id=parent)
        return q.insert(inserted, parent_id=parent_operator_node.name)

    # compound
    def bool(self, *args, **kwargs):
        return self._compound_insert("bool", *args, **kwargs)

    def boost(self, *args, **kwargs):
        return self._compound_insert("boosting", *args, **kwargs)

    def constant_score(self, *args, **kwargs):
        return self._compound_insert("constant_score", *args, **kwargs)

    def dis_max(self, *args, **kwargs):
        return self._compound_insert("dis_max", *args, **kwargs)

    def function_score(self, *args, **kwargs):
        return self._compound_insert("function_score", *args, **kwargs)

    def nested(self, *args, **kwargs):
        return self._compound_insert("nested", *args, **kwargs)

    def has_child(self, *args, **kwargs):
        return self._compound_insert("has_child", *args, **kwargs)

    def has_parent(self, *args, **kwargs):
        return self._compound_insert("has_parent", *args, **kwargs)

    def parent_id(self, *args, **kwargs):
        return self._compound_insert("parent_id", *args, **kwargs)

    def script_score(self, *args, **kwargs):
        return self._compound_insert("script_score", *args, **kwargs)

    def pinned_query(self, *args, **kwargs):
        return self._compound_insert("pinned_query", *args, **kwargs)

    # compound parameters
    def must(self, *args, **kwargs):
        return self._compound_param_insert("bool", "must", *args, **kwargs)

    def should(self, *args, **kwargs):
        return self._compound_param_insert("bool", "should", *args, **kwargs)

    def must_not(self, *args, **kwargs):
        return self._compound_param_insert("bool", "must_not", *args, **kwargs)

    def filter(self, *args, **kwargs):
        return self._compound_param_insert("bool", "filter", *args, **kwargs)

    def _compound_insert(self, compound_key, *args, **kwargs):
        mode = kwargs.pop("mode", ADD)
        # provided parent is compound, real one is parameter
        parent = kwargs.pop("parent", None)
        parent_param = kwargs.pop("parent_param", None)
        child = kwargs.pop("child", None)
        child_param = kwargs.pop("child_param", None)
        compound_q = self.get_dsl_class(compound_key)(*args, **kwargs)
        return self._insert_into(
            compound_q,
            mode=mode,
            parent=parent,
            parent_param=parent_param,
            child=child,
            child_param=child_param,
        )

    def _compound_param_insert(self, method_name, param_key, *args, **kwargs):
        """Must accept:
        bool, must, terms, some_field=2
        bool, must, Terms()
        bool, must, [Terms()]
        bool, must, {"terms": {"some_field": 2}}
        bool, must, [{"terms": {"some_field": 2}}]
        """
        mode = kwargs.pop("mode", ADD)
        _name = kwargs.pop("_name", None)
        parent = kwargs.pop("parent", None)
        parent_param = kwargs.pop("parent_param", None)
        child = kwargs.pop("child", None)
        child_param = kwargs.pop("child_param", None)

        if len(args) < 1:
            raise ValueError("Invalid: %s %s" % (args, kwargs))

        # special syntax
        if kwargs:
            # only allowed when using special syntax:
            # q.must("term", my__field=23)
            if len(args) > 1:
                raise ValueError("Invalid: %s %s" % (args, kwargs))
            arg = args[0]
            if not isinstance(arg, string_types):
                raise ValueError()
            children = [self._get_dsl_class_from_tree_or_node(arg, **kwargs)]
        else:
            children = []
            for arg in args:
                # .must([{}, {}])
                # .must({}, {})
                children.extend(arg if isinstance(arg, list) else [arg])

        return getattr(self, method_name)(
            mode=mode,
            _name=_name,
            parent=parent,
            parent_param=parent_param,
            child=child,
            child_param=child_param,
            **{param_key: children}
        )

    def _compound_update(self, name, new_compound, mode):
        """Update existing compound clause <name> inplace in query by merging it with provided <new_compound> query.

        Three modes are available:
        Mode 'add':
        >>> initial_compound = Query({'bool': {'must': [A, B]}, '_name': 'some_bool_id'})
        >>> new_compound = Query({'bool': {'must': [C], 'filter': [D]}, '_name': 'some_bool_id'})

        :name: bool name in current query
        :param new_compound:
        :param mode: 'add', 'replace' or 'replace_all'.
        :return: self
        """
        if mode not in (ADD, REPLACE, REPLACE_ALL):
            raise ValueError("Unsupported mode <%s> to update compound clause" % mode)
        parent_node = self.parent(name, id_only=False)
        if parent_node is None:
            parent = None
        else:
            parent = parent_node.identifier

        if mode == REPLACE_ALL:
            self.drop_subtree(name)
            self.insert(new_compound, parent_id=parent)
            return self

        for param_node in new_compound.children(new_compound.root, id_only=False):
            existing_param = next(
                (
                    p
                    for p in self.children(name, id_only=False)
                    if p.KEY == param_node.KEY
                ),
                None,
            )
            if not existing_param:
                self.insert(
                    item=new_compound.subtree(param_node.identifier), parent_id=name,
                )
                continue
            if mode == REPLACE:
                self.drop_node(existing_param.identifier)
                self.insert(
                    item=new_compound.subtree(param_node.identifier), parent_id=name,
                )
                continue
            if mode == ADD:
                for clause_node in new_compound.children(
                    param_node.identifier, id_only=False
                ):
                    self.insert(
                        item=new_compound.subtree(clause_node.identifier),
                        parent_id=existing_param.identifier,
                    )
                continue

        # update simple body parameters
        new_compound_node = new_compound.get(new_compound.root)
        current_root_node = self.get(self.root)
        current_root_node.body.update(new_compound_node.body)
        return self

    def __str__(self):
        return "<Query>\n%s" % text(self.show())


class Leaf(Query):
    KEY = None

    def __init__(self, *args, **kwargs):
        mapping = kwargs.pop("mapping", None)
        nested_autocorrect = kwargs.pop("nested_autocorrect", None)
        super(Leaf, self).__init__(
            mapping=mapping, nested_autocorrect=nested_autocorrect
        )
        node = self.get_node_dsl_class(self.KEY)(*args, **kwargs)
        if not isinstance(node, LeafQueryClause):
            raise ValueError("Error: not a leaf %s" % self.KEY)
        self.insert(node)


class Compound(Query):

    KEY = None
    _params_keys = None

    def __init__(self, **kwargs):
        mapping = kwargs.pop("mapping", None)
        nested_autocorrect = kwargs.pop("nested_autocorrect", None)
        super(Compound, self).__init__(
            mapping=mapping, nested_autocorrect=nested_autocorrect
        )
        body = kwargs.copy()
        extra_body = dict()

        # separate body parameters with children clauses (should), from regular body parameters (minimum_should_match=1)
        node_klass = self.get_node_dsl_class(self.KEY)
        if not issubclass(node_klass, CompoundClause):
            raise ValueError("Error: not a compound clause %s" % self.KEY)
        for p_key in node_klass._parent_params:
            p_value = body.pop(p_key, None)
            if p_value is None:
                continue
            extra_body[p_key] = p_value

        # insert compound node (for instance bool)
        compound_node = node_klass(**body)
        self.insert_node(compound_node)

        for p_key, p_value in iteritems(extra_body):
            # insert parameter (for instance must)
            p_klass = self.get_node_dsl_class(p_key)
            p_node = p_klass()
            self.insert_node(p_node, parent_id=compound_node.identifier)

            # atomize
            # filter=[{}, {}], filter={}, filter=Bool(), filter=[Bool(), Term()]
            p_values = p_value if isinstance(p_value, (list, tuple)) else (p_value,)
            if len(p_values) > 1 and not p_klass.MULTIPLE:
                raise ValueError(
                    "Not possible to have multiple clauses under %s parameter."
                    % p_klass.KEY
                )

            # deserialize
            for v in p_values:
                if isinstance(v, (Query, QueryClause)):
                    self.insert(v, parent_id=p_node.identifier)
                    continue
                if isinstance(v, dict):
                    t = self._from_dict(v)
                    self.insert(t, parent_id=p_node.identifier)
                    continue
                raise ValueError("Unkown clause %s" % v)



from .abstract import Leaf


class DistanceFeature(Leaf):
    KEY = "distance_feature"


class MoreLikeThis(Leaf):
    KEY = "more_like_this"


class Percolate(Leaf):
    KEY = "percolate"


class RankFeature(Leaf):
    KEY = "rank_feature"


class Script(Leaf):
    KEY = "script"


class Wrapper(Leaf):
    KEY = "wrapper"



from pandagg.node.query.compound import CompoundClause


class ScriptScore(CompoundClause):
    _default_operator = "query"
    _parent_params = ["query"]
    KEY = "script_score"


class PinnedQuery(CompoundClause):
    _default_operator = "organic"
    _parent_params = ["organic"]
    KEY = "pinned"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from .abstract import UnnamedField


# Identity meta fields
class Index(UnnamedField):
    """The index to which the document belongs."""

    KEY = "_index"


class Type(UnnamedField):
    """The document’s mapping type."""

    KEY = "_type"


class Id(UnnamedField):
    """The document’s ID."""

    KEY = "_id"


# Document source meta-fields
class Source(UnnamedField):
    """The original JSON representing the body of the document."""

    KEY = "_source"


class Size(UnnamedField):
    """The size of the _source field in bytes, provided by the mapper-size plugin."""

    KEY = "_size"


# Indexing meta-fields
class FieldNames(UnnamedField):
    """All fields in the document which contain non-null values."""

    KEY = "_field_names"


class Ignored(UnnamedField):
    """All fields in the document that have been ignored at index time because of ignore_malformed."""

    KEY = "_ignored"


# Routing meta-field
class Routing(UnnamedField):
    """A custom routing value which routes a document to a particular shard."""

    KEY = "_routing"


# Other meta-field
class Meta(UnnamedField):
    """Application specific metadata."""

    KEY = "_meta"



from .abstract import (
    LeafQueryClause,
    AbstractSingleFieldQueryClause,
    KeyFieldQueryClause,
)


class Exists(LeafQueryClause):
    KEY = "exists"

    def __init__(self, field, _name=None):
        self.field = field
        super(Exists, self).__init__(_name=_name, field=field)

    def line_repr(self, depth, **kwargs):
        return "%s, field=%s" % (self.KEY, self.field)


class Fuzzy(KeyFieldQueryClause):
    KEY = "fuzzy"
    _implicit_param = "value"


class Ids(LeafQueryClause):
    KEY = "ids"

    def __init__(self, values, _name=None):
        self.field = "id"
        self.values = values
        super(Ids, self).__init__(_name=_name, values=values)

    def to_dict(self, with_name=True):
        b = {"values": self.values}
        if with_name and self._named:
            b["_name"] = self.name
        return {self.KEY: b}

    def line_repr(self, depth, **kwargs):
        return "%s, values=%s" % (self.KEY, self.values)


class Prefix(KeyFieldQueryClause):
    KEY = "prefix"
    _implicit_param = "value"


class Range(KeyFieldQueryClause):
    KEY = "range"


class Regexp(KeyFieldQueryClause):
    KEY = "regexp"
    _implicit_param = "value"


class Term(KeyFieldQueryClause):
    KEY = "term"
    _implicit_param = "value"


class Terms(AbstractSingleFieldQueryClause):
    KEY = "terms"

    def __init__(self, **body):
        _name = body.pop("_name", None)
        boost = body.pop("boost", None)
        if len(body) != 1:
            raise ValueError("Wrong declaration: %s" % body)
        field, terms = self.expand__to_dot(body).popitem()
        b = {field: terms}
        if boost is not None:
            b["boost"] = boost
        super(Terms, self).__init__(_name=_name, field=field, **b)


class TermsSet(KeyFieldQueryClause):
    KEY = "terms_set"
    _implicit_param = "terms"


class Type(KeyFieldQueryClause):
    KEY = "type"
    _implicit_param = "value"


class Wildcard(KeyFieldQueryClause):
    KEY = "wildcard"
    _implicit_param = "value"



from pandagg.tree.aggs.aggs import Aggs

from pandagg.tree.aggs.bucket import (
    Terms,
    Filters,
    Histogram,
    DateHistogram,
    Global,
    Filter,
    Nested,
    ReverseNested,
    Range,
    Missing,
)

from pandagg.tree.aggs.metric import (
    Avg,
    Max,
    Min,
    Sum,
    Cardinality,
    Stats,
    ExtendedStats,
    Percentiles,
    PercentileRanks,
    GeoBound,
    GeoCentroid,
    TopHits,
    ValueCount,
)

from pandagg.tree.aggs.pipeline import (
    AvgBucket,
    Derivative,
    MaxBucket,
    MinBucket,
    SumBucket,
    StatsBucket,
    ExtendedStatsBucket,
    PercentilesBucket,
    MovingAvg,
    CumulativeSum,
    BucketScript,
    BucketSelector,
    BucketSort,
    SerialDiff,
)

__all__ = [
    "Aggs",
    "Terms",
    "Filters",
    "Histogram",
    "DateHistogram",
    "Range",
    "Global",
    "Filter",
    "Missing",
    "Nested",
    "ReverseNested",
    "Avg",
    "Max",
    "Sum",
    "Min",
    "Cardinality",
    "Stats",
    "ExtendedStats",
    "Percentiles",
    "PercentileRanks",
    "GeoBound",
    "GeoCentroid",
    "TopHits",
    "ValueCount",
    "AvgBucket",
    "Derivative",
    "MaxBucket",
    "MinBucket",
    "SumBucket",
    "StatsBucket",
    "ExtendedStatsBucket",
    "PercentilesBucket",
    "MovingAvg",
    "CumulativeSum",
    "BucketScript",
    "BucketSelector",
    "BucketSort",
    "SerialDiff",
]



import json
from os.path import join
from elasticsearch import Elasticsearch, helpers
from examples.imdb.conf import ES_HOST, DATA_DIR
from pandagg.mapping import (
    Mapping,
    Keyword,
    Text,
    Float,
    Nested,
    Integer,
)

index_name = "movies"
mapping = Mapping(
    properties={
        "movie_id": Keyword(),
        "name": Text(fields={"raw": Keyword()}),
        "year": Integer(),
        "rank": Float(),
        "genres": Keyword(),
        "roles": Nested(
            properties={
                "role": Keyword(),
                "actor_id": Keyword(),
                "gender": Keyword(),
                "first_name": Text(fields={"raw": Keyword()}),
                "last_name": Text(fields={"raw": Keyword()}),
                "full_name": Text(fields={"raw": Keyword()}),
            }
        ),
        "directors": Nested(
            properties={
                "director_id": Keyword(),
                "first_name": Text(fields={"raw": Keyword()}),
                "last_name": Text(fields={"raw": Keyword()}),
                "full_name": Text(fields={"raw": Keyword()}),
                "genres": Keyword(),
            }
        ),
        "nb_directors": Integer(),
        "nb_roles": Integer(),
    }
).to_dict()


def bulk_index(client, docs):
    helpers.bulk(
        client=client,
        actions=[
            {
                "_index": index_name,
                "_op_type": "index",
                "_id": document["movie_id"],
                "_source": document,
            }
            for document in docs
        ],
    )


if __name__ == "__main__":
    es_client = Elasticsearch(hosts=[ES_HOST])

    if es_client.indices.exists(index=index_name):
        print("-" * 50)
        print("DELETE INDEX\n")
        es_client.indices.delete(index=index_name)

    print("-" * 50)
    print("CREATE INDEX\n")
    es_client.indices.create(index_name)
    print("-" * 50)
    print("UPDATE MAPPING\n")
    es_client.indices.put_mapping(index=index_name, body=mapping)

    print("-" * 50)
    print("WRITE DOCUMENTS\n")
    docs_buffer = []
    with open(join(DATA_DIR, "serialized.json"), "r") as f:
        for l in f.readlines():
            if len(docs_buffer) >= 100:
                bulk_index(es_client, docs_buffer)
                docs_buffer = []
            s = json.loads(l)
            docs_buffer.append(s)
    if docs_buffer:
        bulk_index(es_client, docs_buffer)

    es_client.indices.refresh(index=index_name)



from collections import OrderedDict
from unittest import TestCase
from mock import Mock, patch

from pandagg.tree.aggs.aggs import Aggs
from pandagg.tree.response import AggsResponseTree
from pandagg.interactive.response import IResponse
from pandagg.utils import equal_queries
from tests.testing_samples.mapping_example import MAPPING
import tests.testing_samples.data_sample as sample


class ResponseTestCase(TestCase):
    @patch("uuid.uuid4")
    def test_response_tree(self, uuid_mock):
        uuid_mock.side_effect = range(1000)
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response_tree = AggsResponseTree(aggs=my_agg, index=None).parse(
            sample.ES_AGG_RESPONSE
        )
        self.assertEqual(response_tree.__str__(), sample.EXPECTED_RESPONSE_TREE_REPR)
        self.assertEqual(len(response_tree.list()), 18)

        multilabel_gpc_bucket = next(
            (
                b
                for b in response_tree.list()
                if b.level == "global_metrics.field.name" and b.key == "gpc"
            )
        )

        # bucket properties will give parents levels and keys
        self.assertEqual(
            response_tree.bucket_properties(multilabel_gpc_bucket),
            OrderedDict(
                [
                    ("global_metrics.field.name", "gpc"),
                    ("classification_type", "multilabel"),
                ]
            ),
        )


class ClientBoundResponseTestCase(TestCase):
    @patch("lighttree.node.uuid.uuid4")
    def test_client_bound_response(self, uuid_mock):
        uuid_mock.side_effect = range(1000)
        client_mock = Mock(spec=["search"])

        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response_tree = AggsResponseTree(aggs=my_agg, index=None).parse(
            sample.ES_AGG_RESPONSE
        )

        response = IResponse(
            client=client_mock,
            tree=response_tree,
            index_name="some_index",
            depth=1,
            query={"term": {"some_field": 1}},
        )

        # ensure that navigation to attributes works with autocompletion (dir is used in ipython)
        self.assertIn("classification_type_multiclass", dir(response))
        self.assertIn("classification_type_multilabel", dir(response))

        multilabel = response.classification_type_multilabel
        self.assertIsInstance(multilabel, IResponse)
        self.assertIs(multilabel._initial_tree, response._tree)

        self.assertIn("global_metrics_field_name_gpc", dir(multilabel))
        gpc = multilabel.global_metrics_field_name_gpc
        self.assertIsInstance(gpc, IResponse)
        self.assertIs(gpc._initial_tree, response._tree)

        # test filter query used to list documents belonging to bucket
        self.assertTrue(
            equal_queries(
                gpc.get_bucket_filter(),
                {
                    "bool": {
                        "must": [
                            {"term": {"global_metrics.field.name": {"value": "gpc"}}},
                            {"term": {"classification_type": {"value": "multilabel"}}},
                            {"term": {"some_field": {"value": 1}}},
                        ]
                    }
                },
            )
        )
